	!--------------------------------------------------------------------
	!
	!
	! PURPOSE
	!
	! This program solves nonlinear Schrodinger equation in 2 dimensions
	! i*u_t+Es*|u|^2u+u_{xx}+u_{yy}=0
	! using a second order time spectral splitting scheme
	!
	! The boundary conditions are u(x=0,y)=u(2*Lx*\pi,y), 
	!	u(x,y=0)=u(x,y=2*Ly*\pi)
	! The initial condition is u=exp(-x^2-y^2)
	!
	! AUTHORS
	!
	! B. Cloutier, B.K. Muite, P. Rigge
	! 4 June 2012
	!
	! .. Parameters ..
	!  Nx				= number of modes in x - power of 2 for FFT
	!  Ny				= number of modes in y - power of 2 for FFT
	!  Nt				= number of timesteps to take
	!  Tmax				= maximum simulation time
	!  plotgap			= number of timesteps between plots
	!  FFTW_IN_PLACE 	= value for FFTW input 
	!  FFTW_MEASURE 	= value for FFTW input
	!  FFTW_EXHAUSTIVE 	= value for FFTW input
	!  FFTW_PATIENT 	= value for FFTW input    
	!  FFTW_ESTIMATE 	= value for FFTW input
	!  FFTW_FORWARD     = value for FFTW input
	!  FFTW_BACKWARD	= value for FFTW input	
	!  pi = 3.14159265358979323846264338327950288419716939937510d0
	!  Lx				= width of box in x direction
	!  Ly				= width of box in y direction
	!  ES				= +1 for focusing and -1 for defocusing
	! .. Scalars ..
	!  i				= loop counter in x direction
	!  j				= loop counter in y direction
	!  n				= loop counter for timesteps direction	
	!  allocatestatus	= error indicator during allocation
	!  numthreads		= number of openmp threads
	!  ierr				= error return code
	!  start			= variable to record start time of program
	!  finish			= variable to record end time of program
	!  count_rate		= variable for clock count rate
	!  planfxy			= Forward fft plan
	!  planbxy			= Backward fft plan
	!  dt				= timestep
	! .. Arrays ..
	!  unax 			= temporary field
	!  vnax 			= temporary field
	!  potx 			= potential
	! .. Vectors ..
	!  kx				= fourier frequencies in x direction
	!  ky				= fourier frequencies in y direction
	!  x				= x locations
	!  y				= y locations
	!  time				= times at which save data
	!  name_config		= array to store filename for data to be saved    		
	!
	! REFERENCES
	!
	! ACKNOWLEDGEMENTS
	!
	! ACCURACY
	!		
	! ERROR INDICATORS AND WARNINGS
	!
	! FURTHER COMMENTS
	! Check that the initial iterate is consistent with the 
	! boundary conditions for the domain specified
	!--------------------------------------------------------------------
	! External routines required
	! 
	! External libraries required
	! FFTW3	 -- Fast Fourier Transform in the West Library
	!			(http://www.fftw.org/)
	! OpenMP library
		
	PROGRAM main
	USE omp_lib		 	   
	! Declare variables
	IMPLICIT NONE					 
	INTEGER(kind=4), PARAMETER 	::  Nx=8192
	INTEGER(kind=4), PARAMETER 	::  Ny=8192	
	INTEGER(kind=4), PARAMETER	::  Nt=20		
	INTEGER(kind=4), PARAMETER	::  plotgap=5	
	REAL(kind=8), PARAMETER		::  &
	pi=3.14159265358979323846264338327950288419716939937510d0
	REAL(kind=8), PARAMETER		::  Lx=5.0d0		
	REAL(kind=8), PARAMETER		::  Ly=5.0d0		
	REAL(kind=8), PARAMETER		::  Es=1.0d0		
	REAL(kind=8)				::  dt=0.010d0/Nt	
	COMPLEX(kind=8), DIMENSION(:), ALLOCATABLE	::  kx			
	COMPLEX(kind=8), DIMENSION(:), ALLOCATABLE	::  ky			
	REAL(kind=8),  	 DIMENSION(:), ALLOCATABLE	::  x			
	REAL(kind=8),  	 DIMENSION(:), ALLOCATABLE	::  y			
	COMPLEX(kind=8), DIMENSION(:,:), ALLOCATABLE::  unax,vnax,potx 
	REAL(kind=8), 	 DIMENSION(:), ALLOCATABLE	::  time
	REAL(kind=8)				:: scalemodes
	INTEGER(kind=4)				::  i,j,k,n,allocatestatus,ierr
	INTEGER(kind=4)			 	::  start, finish, count_rate, numthreads
	INTEGER(kind=8), PARAMETER  ::  FFTW_IN_PLACE=8, FFTW_MEASURE=0,&
									FFTW_EXHAUSTIVE=8, FFTW_PATIENT=32,&
                   	 				FFTW_ESTIMATE=64                                   
	INTEGER(kind=8),PARAMETER   ::  FFTW_FORWARD=-1, FFTW_BACKWARD=1	
	INTEGER(kind=8)				::  planfxy,planbxy
   	CHARACTER*100			 	::  name_config
	
	numthreads=omp_get_max_threads() 
	PRINT *,'Program starting'
	PRINT *,'Grid: ',Nx,'X',Ny	
	PRINT *,'There are ',numthreads,' threads.'
	    			   	
	ALLOCATE(kx(1:Nx),ky(1:Nx),x(1:Nx),y(1:Nx),unax(1:Nx,1:Ny),&
			vnax(1:Nx,1:Ny),potx(1:Nx,1:Ny),time(1:1+Nt/plotgap),&
			stat=allocatestatus)	
	IF (allocatestatus .ne. 0) stop 
	PRINT *,'allocated memory'

	! set up multithreaded ffts
	CALL dfftw_init_threads_(ierr)
	PRINT *,'Initiated threaded FFTW'
	CALL dfftw_plan_with_nthreads_(numthreads)
	PRINT *,'Inidicated number of threads to be used in planning'
	CALL dfftw_plan_dft_2d_(planfxy,Nx,Ny,unax(1:Nx,1:Ny),vnax(1:Nx,1:Ny),&
							FFTW_FORWARD,FFTW_ESTIMATE)
 	CALL dfftw_plan_dft_2d_(planbxy,Nx,Ny,vnax(1:Nx,1:Ny),unax(1:Nx,1:Ny),&
 							FFTW_BACKWARD,FFTW_ESTIMATE)
	PRINT *,'Setup FFTs'
	scalemodes=1.0d0/REAL(Nx*Ny,kind(0d0))	
	! setup fourier frequencies
	!$OMP PARALLEL PRIVATE(i,j) 
	!$OMP DO SCHEDULE(static)
	DO i=1,1+Nx/2
		kx(i)= cmplx(0.0d0,1.0d0)*REAL(i-1,kind(0d0))/Lx  			
	END DO
	!$OMP END DO
	kx(1+Nx/2)=0.0d0
	!$OMP DO SCHEDULE(static)
	DO i = 1,Nx/2 -1
		kx(i+1+Nx/2)=-kx(1-i+Nx/2)
	END DO
	!$OMP END DO
	!$OMP DO SCHEDULE(static)
  	DO i=1,Nx
		x(i)=(-1.0d0+2.0d0*REAL(i-1,kind(0d0))/REAL(Nx,kind(0d0)) )*pi*Lx
	END DO
	!$OMP END DO
	!$OMP DO SCHEDULE(static)
	DO j=1,1+Ny/2
		ky(j)= cmplx(0.0d0,1.0d0)*REAL(j-1,kind(0d0))/Ly  			
	END DO
	!$OMP END DO
	ky(1+Ny/2)=0.0d0
	!$OMP DO SCHEDULE(static)
	DO j = 1,Ny/2 -1
		ky(j+1+Ny/2)=-ky(1-j+Ny/2)
	END DO
	!$OMP END DO
	!$OMP DO SCHEDULE(static)
  	DO j=1,Ny
		y(j)=(-1.0d0+2.0d0*REAL(j-1,kind(0d0))/REAL(Ny,kind(0d0)) )*pi*Ly
	END DO
	!$OMP END DO
	PRINT *,'Setup grid and fourier frequencies'
	!$OMP DO SCHEDULE(static)
	DO j=1,Ny
		DO i=1,Nx
			unax(i,j)=exp(-1.0d0*(x(i)**2 +y(j)**2)) 
		END DO
	END DO
	!$OMP END DO
	!$OMP END PARALLEL
	name_config = 'uinitial.dat' 
	OPEN(unit=11,FILE=name_config,status="UNKNOWN") 	
	REWIND(11)
	DO j=1,Ny
		DO i=1,Nx
			WRITE(11,*) abs(unax(i,j))**2
		END DO
	END DO
	CLOSE(11)
	CALL system_clock(start,count_rate)	

	CALL dfftw_execute_dft_(planfxy,unax(1:Nx,1:Ny),vnax(1:Nx,1:Ny)) 
	! transform initial data and do first half time step
	!$OMP PARALLEL DO PRIVATE(i,j) SCHEDULE(static)
	DO j=1,Ny
		DO i=1,Nx
			vnax(i,j)=exp(0.5d0*dt*(kx(i)*kx(i) + ky(j)*ky(j))&
					*cmplx(0.0d0,1.0d0))*vnax(i,j)
		END DO
	END DO
	!$OMP END PARALLEL DO
	PRINT *,'Got initial data, starting timestepping'
	time(1)=0.0d0
	DO n=1,Nt					
		CALL dfftw_execute_dft_(planbxy,vnax(1:Nx,1:Ny),unax(1:Nx,1:Ny))
		!$OMP PARALLEL PRIVATE(i,j) 
		!$OMP DO SCHEDULE(static)
		DO j=1,Ny
			DO i=1,Nx
				potx(i,j)=Es*unax(i,j)*conjg(unax(i,j))*scalemodes*scalemodes
			END DO
		END DO
		!$OMP END DO 
		!$OMP DO SCHEDULE(static)
		DO j=1,Ny
			DO i=1,Nx
				unax(i,j)=exp(cmplx(0.0d0,-1.0d0)*dt*potx(i,j))&
						*unax(i,j)*scalemodes
			END DO
		END DO
		!$OMP END DO
		!$OMP END PARALLEL  
		CALL dfftw_execute_dft_(planfxy,unax(1:Nx,1:Ny),vnax(1:Nx,1:Ny))
		!$OMP PARALLEL DO PRIVATE(i,j) SCHEDULE(static)
		DO j=1,Ny
			DO i=1,Nx
				vnax(i,j)=exp(dt*(kx(i)*kx(i) + ky(j)*ky(j))&	
						*cmplx(0.0d0,1.0d0))*vnax(i,j)
			END DO	
		END DO
		!$OMP END PARALLEL DO
		IF (mod(n,plotgap)==0) then
			time(1+n/plotgap)=n*dt
			PRINT *,'time',n*dt
		END IF
	END DO		
		
	! transform back final data and do another half time step
	CALL dfftw_execute_dft_(planbxy,vnax(1:Nx,1:Ny),unax(1:Nx,1:Ny))
	!$OMP PARALLEL PRIVATE(i,j) 
	!$OMP DO SCHEDULE(static)
	DO j=1,Ny
		DO i=1,Nx
			potx(i,j)=Es*unax(i,j)*conjg(unax(i,j))*scalemodes*scalemodes
		END DO
	END DO
	!$OMP END DO
	!$OMP DO SCHEDULE(static)
	DO j=1,Ny
		DO i=1,Nx
			unax(i,j)=exp(cmplx(0,-1)*dt*potx(i,j))*unax(i,j)*scalemodes
		END DO
	END DO
	!$OMP END DO
	!$OMP END PARALLEL
	CALL dfftw_execute_dft_(planfxy,unax(1:Nx,1:Ny),vnax(1:Nx,1:Ny))
	!$OMP PARALLEL DO PRIVATE(i,j) SCHEDULE(static)
	DO j=1,Ny
		DO i=1,Nx
			vnax(i,j)=exp(0.5d0*dt*(kx(i)*kx(i) + ky(j)*ky(j))&
					*cmplx(0.0d0,1.0d0))*vnax(i,j)*scalemodes	
		END DO	
	END DO
	!$OMP END PARALLEL DO
	CALL dfftw_execute_dft_(planbxy,vnax(1:Nx,1:Ny),unax(1:Nx,1:Ny))
	PRINT *,'Finished time stepping'
	CALL system_clock(finish,count_rate)
	PRINT*,'Program took ',REAL(finish-start)/REAL(count_rate),&
		'for Time stepping'

	name_config = 'ufinal.dat' 
	OPEN(unit=11,FILE=name_config,status="UNKNOWN") 	
	REWIND(11)
	DO j=1,Ny
		DO i=1,Nx
			WRITE(11,*) abs(unax(i,j))**2
		END DO
	END DO
	CLOSE(11)
		
	name_config = 'tdata.dat' 
	OPEN(unit=11,FILE=name_config,status="UNKNOWN") 	
	REWIND(11)
	DO j=1,1+Nt/plotgap
		WRITE(11,*) time(j)
	END DO
	CLOSE(11)

	name_config = 'xcoord.dat' 
	OPEN(unit=11,FILE=name_config,status="UNKNOWN") 	
	REWIND(11)
	DO i=1,Nx
		WRITE(11,*) x(i)
	END DO
	CLOSE(11)

	name_config = 'ycoord.dat' 
	OPEN(unit=11,FILE=name_config,status="UNKNOWN") 	
	REWIND(11)
	DO j=1,Ny
		WRITE(11,*) y(j)
	END DO
	CLOSE(11)
	PRINT *,'Saved data'

	CALL dfftw_destroy_plan_(planbxy)
	CALL dfftw_destroy_plan_(planfxy)
	CALL dfftw_cleanup_threads_()

	DEALLOCATE(unax,vnax,potx,time,kx,ky,x,y,stat=allocatestatus)	
	IF (allocatestatus .ne. 0) STOP 
	PRINT *,'Deallocated memory'
		
   	PRINT *,'Program execution complete'
	END PROGRAM main
