	SUBROUTINE computu(	Nx, Ny, n_p, t_p, xdiff, ydiff, info_p, &
					PLAN2X, PLAN2Y, PLAN1X, PLAN1Y, lowxv, upxv, lowyv, upyv, &
					UINI, VINI, U2B, AUINI,T1,T2, W, DT, EPSI, &
					Kx, Kxb, Ky, RHO )  
	!---------------------------------------------------------------------------------------------
	! .. PURPOSE ..
	!	computu solves the equation 
	!	u_t=-\rho/i*epsi*(abs(u)^2)*u 
	!	in physical space, since abs(u)^2 is time independant
	!
	! .. SPECIFICATION ..
	!
	!	computu( Nx, Ny, n_p, t_p, xdiff, ydiff, info_p, &
	!		PLAN2X, PLAN2Y, PLAN1X, PLAN1Y, lowxv, upxv, lowyv, upyv, &
	!		UINI, VINI, U2B, AUINI,T1,T2, W, DT, EPSI, &
	!		Kx, Kxb, Ky, RHO )  
	!
	!	INTEGER(kind=4)	lowxv(0:n_p-1), upxv(0:n_p-1), lowyv(0:n_p-1), upyv(0:n_p-1), &
	!			Nx, Ny, n_p, t_p, xdiff, ydiff, info_p
	!	INTEGER(kind=8)	PLAN2X, PLAN2Y, PLAN1X, PLAN1Y
	!	DOUBLE PRECISION	U2B(1:Nx, lowyv(t_p):upyv(t_p)), Kx(lowxv(t_p):upxv(t_p)), &
	!				Kxb(lowxv(t_p):upxv(t_p)), Ky(1:Ny), W, DT, EPSI, RHO
	!	COMPLEX(kind=8)	AUINI(1:Nx, lowyv(t_p):upyv(t_p)), UINI(1:Nx, lowyv(t_p):upyv(t_p)), &
	!			T1(1:ydiff*Nx), T2(1:ydiff*Nx), VINI(1:Ny, lowxv(t_p):upxv(t_p))
	!
	! .. INPUT ..
	! Ny		= The number of grid points in the vertical direction 
	! Nx		= The number of grid points in the horizontal direction
	! info_p	= MPI error return integer
	! n_p		= Number of MPI processes
	! t_p		= Number of this MPI process
	! lowyv		= lower array index for data to be saved to file
	! upyv		= upper array index for data to be saved to file
	! lowxv		= lower array index for data to be saved to file
	! upxv		= upper array index for data to be saved to file
	! RHO		= +1 for the defocusing case and -1 for the focusing case
	! PLAN2X	= Backward fft in X direction
	! PLAN2Y	= Backward fft in Y direction
	! PLAN1X	= Forward fft in X direction
	! PLAN1Y	= Forward fft in Y direction
	! Kx		= X fourier frequencies
	! Kxb		= X fourier frequencies, modified so do not get division by zero
	! Ky		= Y fourier frequencies
	! W		= Weight for splitting
	! DT		= Time step
	! EPSI		= dispersion parameter, 0 < EPSI << 1
	! UINI		= u, solution of DS, size: [Nx Ny/n_p]
	! VINI 		= FFT(u), (FFT in both spatial coordinates), size: [Ny Nx/n_p]
	! AUINI		= Absolute Value of u, squared; size: [Nx Ny/n_p]
	! U2B		= -((abs(u).^2) + IFFT(K.*FFT(abs(u).^2))); SIZE: [Nx Ny/n_p]
	!		where, IFFT denotes the backward FFT in both spatial coordinates,
	!		and K denotes the operator with which we solve the second equation in
	!		DS, the equation for \phi. (K=(-Kx.^2)./(Kx.^2 + ALPHA*Ky.^2).
	! T1		= swap space for FFT
	! T2		= swap space for FFT
	!
	! REFERENCES
	!
	! ACCURACY
	!		
	! ERROR INDICATORS AND WARNINGS
	!
	! FURTHER COMMENTS
	!---------------------------------------------------------------------------------------------
	! EXTERNAL ROUTINES REQUIRED
	! MPI library
	!---------------------------------------------------------------------------------------------

	USE mpi      
	IMPLICIT NONE    
	! .. Scalar Arguments ..
	INTEGER(KIND=4), INTENT(IN)						::  Nx, Ny ,n_p, t_p, xdiff, ydiff
	INTEGER(KIND=4), INTENT(INOUT)					::  info_p
	INTEGER(KIND=8), INTENT(INOUT)					::  PLAN2X, PLAN2Y, PLAN1X, PLAN1Y
	DOUBLE PRECISION, INTENT(IN)					::  W, DT, EPSI, RHO
	! .. Array arguments ..
	INTEGER(KIND=4), DIMENSION(0:n_p-1), INTENT(IN)				::  lowxv, upxv, lowyv, upyv
	DOUBLE PRECISION, DIMENSION(lowxv(t_p):upxv(t_p)), INTENT(IN)		::  Kx, Kxb  
	DOUBLE PRECISION, DIMENSION(1:Ny), INTENT(IN)						::  Ky 
	DOUBLE PRECISION, DIMENSION(1:Nx,lowyv(t_p):upyv(t_p)), INTENT(OUT)	::  U2B
	COMPLEX(KIND=8), DIMENSION(1:Nx,lowyv(t_p):upyv(t_p)), INTENT(INOUT)::  UINI
	COMPLEX(KIND=8), DIMENSION(1:Nx,lowyv(t_p):upyv(t_p)), INTENT(OUT)	::  AUINI
	COMPLEX(KIND=8), DIMENSION(1:Ny,lowxv(t_p):upxv(t_p)), INTENT(OUT)	::  VINI
	COMPLEX(KIND=8), DIMENSION(1:ydiff*Nx), INTENT(OUT)					::  T1, T2
	! .. Local Scalars ..
	INTEGER(KIND=4)														::  i,j
	! .. Executable statements ..
	CONTINUE

	DO j = lowyv(t_p),upyv(t_p)
		AUINI(1:Nx,j) = ABS(UINI(1:Nx,j))**2
	END DO		
	DO j=lowyv(t_p),upyv(t_p)
		UINI(1:Nx,j) = UINI(1:Nx,j)*EXP( RHO*W*CMPLX(0d0,-1.0d0)*AUINI(1:Nx,j)*DT/EPSI)
	END DO 

	END SUBROUTINE computu
