	SUBROUTINE computv(   Nx, Ny, n_p, t_p, xdiff, ydiff, info_p, &
				   PLAN2X, PLAN2Y, PLAN1X, PLAN1Y, lowxv, upxv, lowyv, upyv, &
				   UINI, VINI, T1, T2, W, DT, EPSI, Kx, Ky )  
	!---------------------------------------------------------------------------------------------
	! .. PURPOSE ..
	!	computv solves the equation u_t=i*epsi*(u_xx + u_yy) in Fourier space
	!
	! .. SPECIFICATION ..
	!	computation2(	Nx, Ny, n_p, t_p, xdiff, ydiff, info_p, &
	!			PLAN2X, PLAN2Y, PLAN1X, PLAN1Y, lowxv, upxv, lowyv, upyv, &
	!			UINI, VINI, W, DT, EPSI, Kx, Ky ) 
	!
	!	INTEGER(kind=4)	lowxv(0:n_p-1), upxv(0:n_p-1), lowyv(0:n_p-1), upyv(0:n_p-1), &
	!			Nx, Ny, n_p, t_p, xdiff, ydiff, info_p
	!	INTEGER(kind=8)	PLAN2X, PLAN2Y, PLAN1X, PLAN1Y
	!	DOUBLE PRECISION	LAVINI(1:NX,lowyv:upyv),Kx(lowxv(t_p):upxv(t_p)), &
	!				Ky(1:Ny), W, DT, EPSI
	!	COMPLEX(kind=8)	UINI(1:Nx, lowyv(t_p):upyv(t_p)), VINI(1:Ny, lowxv(t_p):upxv(t_p)), &
	!			T1(1:ydiff*Nx), T2(1:ydiff*Nx)
	!
	! .. INPUT ..
	! Ny		= The number of grid points in the vertical direction 
	! Nx		= The number of grid points in the horizontal direction
	! info_p	= MPI error return integer
	! n_p		= Number of MPI processes
	! t_p		= Number of this MPI process
	! lowyv		= lower array index for data to be saved to file
	! upyv		= upper array index for data to be saved to file
	! lowxv		= lower array index for data to be saved to file
	! upxv		= upper array index for data to be saved to file
	! PLAN2X	= Backward fft in X direction
	! PLAN2Y	= Backward fft in Y direction
	! PLAN1X	= Forward fft in X direction
	! PLAN1Y	= Forward fft in Y direction
	! Kx		= X fourier frequencies
	! Ky		= Y fourier frequencies
	! W		= Weight for splitting
	! DT		= Time step
	! EPSI		= dispersion parameter, 0 < EPSI << 1
	! UINI		= u, solution of DS, size: [Nx Ny/n_p]
	! VINI 		= FFT(u), (FFT in both spatial coordinates), size: [Ny Nx/n_p]
	! T1		= swap space for FFT
	! T2		= swap space for FFT
	!
	! REFERENCES
	!
	! ACCURACY
	!		
	! ERROR INDICATORS AND WARNINGS
	!
	! FURTHER COMMENTS
	!---------------------------------------------------------------------------------------------
	! EXTERNAL ROUTINES REQUIRED
	! MPI library
	! fft2dFor	-- Computes the 2d-FFT forward in both spatial coordinates, by using 
	!		MPI_ALL_TO_ALL to perform the data transposition. Data is only transposed once,
	!		so output is in the opposite orientation to input.  Input data is destroyed.
	! fft2dBack	-- Computes the 2d-FFT backward in both spatial coordinates, by using 
	!		MPI_ALL_TO_ALL to perform the data transposition. Data is only transposed once,
	!		so output is in the opposite orientation to input.  Input data is destroyed.

	USE mpi
	IMPLICIT NONE    
	! .. Scalar Arguments ..
	INTEGER(KIND=4), INTENT(IN)						::  Nx, Ny, n_p, t_p, xdiff, ydiff 
	INTEGER(KIND=4), INTENT(INOUT)					::  info_p
	INTEGER(KIND=8), INTENT(INOUT)					::  PLAN2X, PLAN2Y, PLAN1X, PLAN1Y
	DOUBLE PRECISION, INTENT(IN)					::  W, DT, EPSI
	! .. Array Arguments
	INTEGER(KIND=4), DIMENSION(0:n_p-1), INTENT(IN)	::  lowxv, upxv, lowyv, upyv
	DOUBLE PRECISION, DIMENSION(lowxv(t_p):upxv(t_p)), INTENT(IN)	::  Kx 
	DOUBLE PRECISION, DIMENSION(1:Ny), INTENT(IN)					::  Ky 
	COMPLEX(KIND=8), DIMENSION(1:Nx, lowyv(t_p):upyv(t_p)), INTENT(INOUT)	::  UINI
	COMPLEX(KIND=8), DIMENSION(1:Ny, lowxv(t_p):upxv(t_p)), INTENT(OUT)		::  VINI
	COMPLEX(KIND=8), DIMENSION(1:ydiff*Nx), INTENT(OUT)						::  T1, T2
	! .. Local Scalars
	INTEGER(KIND=4) 														::  i
	! .. Executable statements ..	
	CONTINUE

	CALL fft2dFor(	Nx, Ny, n_p, t_p, xdiff, ydiff, info_p, &
			PLAN1X, PLAN1Y, lowxv(0:n_p-1), upxv(0:n_p-1), lowyv(0:n_p-1), upyv(0:n_p-1), &
			UINI(1:Nx, lowyv(t_p):upyv(t_p)), VINI(1:Ny, lowxv(t_p):upxv(t_p)), &
			T1(1:ydiff*Nx),T2(1:ydiff*Nx))  
	DO i = lowxv(t_p),upxv(t_p)
		VINI(1:Ny,i)=VINI(1:Ny,i)*EXP(W*DT*EPSI*CMPLX(0D0, -1D0)*(Kx(i)*Kx(i) + Ky(1:Ny)*Ky(1:Ny)))
	END DO
	CALL fft2dBack(	Nx, Ny, n_p, t_p, xdiff, ydiff, info_p, &
			PLAN2X, PLAN2Y, lowxv(0:n_p-1), upxv(0:n_p-1), lowyv(0:n_p-1), upyv(0:n_p-1), &
			VINI(1:Ny, lowxv(t_p):upxv(t_p)), UINI(1:Nx, lowyv(t_p):upyv(t_p)), &
			T1(1:ydiff*Nx),T2(1:ydiff*Nx))  

	END SUBROUTINE computv

