	PROGRAM NLS2d
	!-------------------------------------------------------------------------------------
	! .. PURPOSE ..
	!
	!	NLS2d solves the two dimensional Schrodinger equation
	!	(i*epsi*u_t + epsi^2*u_xx + epsi^2*u_yy + 2\rho*(|u|^2)u = 0,
	!	on [-Lx*pi,Lx*pi]x[-Ly*pi,Ly*pi]
	!	by the Time Splitting Method of order 2 
	!
	! .. PARAMETERS .. INITIALIZED IN INPUTFILE
	!	RHO			= +1 for the defocusing case and -1 for the focusing case
	!	Nx			= power of two, number of modes in the x direction
	!	Ny			= power of two, number of modes in the y direction
	!	Nt			= the number of timesteps
	!	plotgap		= the number of timesteps to take before plotting
	!	Lx			= definition of the periodic domain of computation in x direction
	!	Ly			= definition of the periodic domain of computation in y direction
	!	EPSI		= dispersion parameter, 0 < EPSI << 1
	!	DT			= the time step
	!
	! .. Scalars ..
	!	W1, W2, W3, W4		= coefficients of the methods, see REFERENCES
	!	t_p			= the number of the MPI process  
	!	info_p			= MPI error code
	!	n_p			= the total number of MPI processes
	!	i, k, j			= integers used for loops
	!	deb, fin, co		= variables used for the calcul of the CPU time
	!	AllocateStatus		= variable used to verify the allocation of arrays
	!	pi			= the number pi, pi=4*Arctan(1).
	!	plotnum			= number of plots saved
	!	thefileu		= handle for file to be saved to disk
	!	thefilev		= handle for file to be saved to disk
	!
	! .. Variables, 1-dimension .. 	
	!	xdiff			= Nx/n_p  Variable used for the indiciation 
	!				(and so for the partage of datas between processes) 
	!	ydiff			= Ny/n_p  Variable used for the indiciation
	!				(and so for the partage of datas between processes) 
	!	PLAN1X			= data structure containing all the information that 
	!				fftw needs in order to compute the 1D forward Fourier transform
	!				in the x direction
	!	PLAN2X			= data structure containing all the information that 
	!				fftw needs in order to compute the 1D backward Fourier transform
	!				in the x direction
	!	PLAN1Y			= data structure containing all the information that 
	!				fftw needs in order to compute the 1D forward Fourier transform
	!				in the y direction
	!	PLAN2Y			= data structure containing all the information that 
	!				fftw needs in order to compute the 1D backward Fourier transform
	!				in the y direction
	!	MASSCHECK		= test of the numerical precision provided by mass conservation
	!	MA0			= initial mass numerically calculated via FFT
	!				(in both spatial coordinates)
	!
	! .. Arrays 1d .. (Vectors)
	!	COMP1			= (/ W1, W1 /), variable used to compute 
	!				the time splitting method (see REFERENCES) in a single loop 
	!	COMP2			= (/ W2, 0 /), variable used to compute 
	!				the time splitting method (see REFERENCES) in a single loop 
	!	lowxv, upxv, lowyv, upyv = Variables used for the indication, size: n_p
	!				low=lower, up=upper, xv= x value, yv=y value
	!	Kx			= Fourier frequencies in the x direction, size: lowxv(t_p):uppxv(t_p)  
	!	Kxb			= Fourier frequencies in the x direction, modified to avoid division by zero,
	!				size: lowxv(t_p):uppxv(t_p)  
	!	Ky			= Fourier frequencies in the y direction, size: Ny	
	!
	! .. Arrays 2d .. (Matrix)
	!	Note: Sizes are indicated without parallelisation
	!
	!	UINI			= u, solution of DS, size: [Nx Ny/n_p]
	!	VINI			= FFT(u), (FFT in both spatial coordinates), size: [Ny Nx/n_p]
	!	AUINI			= Absolute Value of u, squared; size: [Nx Ny/n_p]
	!	U2B			= -((abs(u).^2) + IFFT(K.*FFT(abs(u).^2))); SIZE: [Nx Ny/n_p]
	!				where, IFFT denotes the backward FFT in both spatial coordinates,
	!				and K denotes the operator with which we solve the second equation in
	!				DS, the equation for \phi. (K=(-2*Kx.^2)./(Kx.^2 + ALPHA*Ky.^2).
	!				Also used as the array to store data to be saved to disk
	! .. REFERENCES ..
	! Yoshida, Haruo: Construction of higher order symplectic integrators, 
	!		Physics Letters A 150,(5-7) pp262-268 November 1990
	! .. ACCURACY ..
	!		
	! .. ERROR INDICATORS AND WARNINGS ..
	! One might possibly use derived datatypes, MPI_ALLTOALL_W and/or change from 
	! storing data in 2d arrays to storing data in large 1d vectors with clever indexing to 
	! reduce memory requirements
	!
	! FURTHER COMMENTS
	! This program is based on a program for the Davey Stewartson equation originally by
	! B.K. Muite and K. Roidot
	!---------------------------------------------------------------------------------------------
	! EXTERNAL ROUTINES REQUIRED
	! MPI library
	! parainit 	-- intialize parameters, which are stocked in the Input File name_file
	! computu 	-- solves the equation 
	!		u_t=-2*\rho/i*epsi*(abs(u)^2)*u 
	!		in physical space, since (abs(u)^2) is time independant
	! computv 	-- solves the equation 
	!		u_t=i*epsi*(u_xx + u_yy) in Fourier space
	! fftinit	-- initializes fftw3 plans
	! writebinfile  -- saves a two dimensional field in native binary, double format using MPIIO
	! fft2dFor  	-- Computes the 2d-FFT forward in both spatial coordinates, by using 
	!		MPI_ALL_TO_ALL to perform the data transposition. Data is only transposed once, so
	!		output is in the opposite orientation to input.  Input data is destroyed.
	! fft2dBack	-- Computes the 2d-FFT backward in both spatial coordinates, by using 
	!		MPI_ALL_TO_ALL to perform the data transposition. Data is only transposed once, so
	!		output is in the opposite orientation to input.  Input data is destroyed.
	! compandsave	-- save the Fourier coefficients and the numerical solution
	!		in a binary file, by calling the subroutine writebinfile,
	!		and calculate the test for the mass conservation.
	! fftdestroy	-- destroys fftw arrays
	! init		-- initializes the grid points of computation, fourier frequencies, 
	!		the initial data, uini, and initial mass, and
	!		also saves initial data required to calculate points.
	!-----------------------------------------------------------------------------------------------------
	USE mpi
      
	IMPLICIT NONE    
!	Declaration of variables and parameter
	INTEGER(KIND=4)					::  Nx, Ny  
	INTEGER(KIND=4)					::  Nt, plotgap
	INTEGER(KIND=4)					::  i, k, j, start, finish, count_rate, AllocateStatus
	INTEGER(KIND=4)					::  info_p, n_p, t_p, xdiff, ydiff, plotnum, thefileu, thefilev
	INTEGER(KIND=8)					::  PLAN1X, PLAN1Y, PLAN2X, PLAN2Y
	DOUBLE PRECISION				::  EPSI
	DOUBLE PRECISION				::  Lx
	DOUBLE PRECISION				::  Ly
	DOUBLE PRECISION, PARAMETER			::  pi= 3.14159265358979323846264338327950288419716939937510
	DOUBLE PRECISION				::  RHO 
	DOUBLE PRECISION, DIMENSION(2), PARAMETER	::  COMP1=(/ 	0.5d0, 0.5d0 /)
	DOUBLE PRECISION, DIMENSION(2), PARAMETER	::  COMP2=(/ 1.0d0, 0.0d0  /)
	DOUBLE PRECISION				::  DT, time, MASSCHECK, MA0  
	INTEGER, DIMENSION(MPI_STATUS_SIZE) 		::  mpistatus
	INTEGER(KIND=4), DIMENSION(:), ALLOCATABLE	::  lowxv, upxv, lowyv, upyv
	DOUBLE PRECISION, DIMENSION(:), ALLOCATABLE	::  Kx, Kxb, Ky 
	COMPLEX(KIND=8), DIMENSION(:), ALLOCATABLE	::  T1, T2
	COMPLEX(KIND=8), DIMENSION(:,:), ALLOCATABLE	::  AUINI, UINI, VINI
	DOUBLE PRECISION, DIMENSION(:,:), ALLOCATABLE	::  U2B, PLOTARRAYU, PLOTARRAYV
!       Initialisation of data 

	
!       Initialisation MPI environment
	CALL MPI_INIT(info_p)
	CALL MPI_COMM_SIZE(MPI_COMM_WORLD, n_p, info_p)
	CALL MPI_COMM_RANK(MPI_COMM_WORLD, t_p, info_p)   
!       give a number to my proc from 0 to n_p-1

!	Initialisation of parameter
	CALL parainit(	Nx, Ny, RHO, Nt, plotgap, Lx, Ly, EPSI, &
			DT, time, t_p, n_p, info_p )

	plotnum=0

!	indiciation 
	ydiff = Ny/n_p
	xdiff = Nx/n_p

	ALLOCATE( lowxv(0:n_p-1) , lowyv(0:n_p-1), upxv(0:n_p-1), upyv(0:n_p-1), stat=AllocateStatus)
	IF (AllocateStatus .NE. 0) STOP

	DO i = 1,n_p
		lowxv(i-1) = (i-1)*xdiff + 1
		lowyv(i-1) = (i-1)*ydiff + 1
		upxv(i-1) = i*xdiff 
		upyv(i-1) = i*ydiff
	END DO
!	end indiciation 

!	allocation of memory
	ALLOCATE(	Kx(lowxv(t_p):upxv(t_p)), Kxb(lowxv(t_p):upxv(t_p)), Ky(1:Ny), &
			UINI(1:Nx,lowyv(t_p):upyv(t_p)), AUINI(1:Nx,lowyv(t_p):upyv(t_p)), &
			U2B(1:Nx,lowyv(t_p):upyv(t_p)), VINI(1:Ny,lowxv(t_p):upxv(t_p)), &
			T1(1:ydiff*Nx),T2(1:ydiff*Nx), PLOTARRAYU(1:Nx,lowyv(t_p):upyv(t_p)),&
			PLOTARRAYV(1:Ny,lowxv(t_p):upxv(t_p)), stat=AllocateStatus   )	 
	IF (AllocateStatus .NE. 0) STOP 

	IF (t_p.EQ.0) THEN
		PRINT *,'allocation memory, OK '
	ENDIF
	! use of fftw needs creation of plans: 
	CALL fftinit( UINI(1:Nx,lowyv(t_p)), VINI(1:Ny,lowxv(t_p)), Nx, Ny, PLAN1X, PLAN1Y, PLAN2X, PLAN2Y )
	IF (t_p.EQ.0) THEN
		PRINT *,'set up fft, OK '
	ENDIF

!	Initialisation of variables, initial data ....
	CALL init(	Nx, Ny, n_p, t_p, xdiff, ydiff, info_p, &
			PLAN1X, PLAN1Y, Lx, Ly, pi, RHO, &
			MA0, lowxv(0:n_p-1), upxv(0:n_p-1), lowyv(0:n_p-1), upyv(0:n_p-1),&
			Kx(lowxv(t_p):upxv(t_p)), Kxb(lowxv(t_p):upxv(t_p)), Ky(1:Ny), &
			UINI(1:Nx,lowyv(t_p):upyv(t_p)), AUINI(1:Nx,lowyv(t_p):upyv(t_p)), &
			VINI(1:Ny,lowxv(t_p):upxv(t_p)), T1(1:ydiff*Nx),T2(1:ydiff*Nx) )
	IF (t_p.EQ.0) THEN
	   PRINT *,'set up initial data '
	ENDIF
				
!		save initial data
	CALL compandsave(	Nx, Ny, n_p, t_p, xdiff, ydiff, info_p, &
				PLAN1X, PLAN1Y, lowxv(0:n_p-1), upxv(0:n_p-1), lowyv(0:n_p-1), upyv(0:n_p-1), &
				UINI(1:Nx,lowyv(t_p):upyv(t_p)), VINI(1:Ny,lowxv(t_p):upxv(t_p)),&
				AUINI(1:Nx,lowyv(t_p):upyv(t_p)), &
				MA0, MASSCHECK, PLOTARRAYU(1:Nx,lowyv(t_p):upyv(t_p)), &
				PLOTARRAYV(1:Ny,lowxv(t_p):upxv(t_p)),T1(1:ydiff*Nx),T2(1:ydiff*Nx),&
				thefileu, thefilev, time, plotnum)
!	end of the initialisation
	
	IF (t_p.EQ.0) THEN
		PRINT *,'saved initial data '
	ENDIF
!	system clock
	call system_clock(start,count_rate)
!	Solve the equation     
!	Start of the loop 
	DO k = 1,Nt
		DO i=1,1
			CALL computv(	Nx, Ny, n_p, t_p, xdiff, ydiff, info_p, &
					PLAN2X, PLAN2Y, PLAN1X, PLAN1Y, &
					lowxv(0:n_p-1), upxv(0:n_p-1), lowyv(0:n_p-1), upyv(0:n_p-1), &
					UINI(1:Nx,lowyv(t_p):upyv(t_p)), VINI(1:Ny,lowxv(t_p):upxv(t_p)), &
					T1(1:ydiff*Nx),T2(1:ydiff*Nx), &
					COMP1(i), DT, EPSI, Kx(lowxv(t_p):upxv(t_p)), Ky(1:Ny) )
			CALL computu(	Nx, Ny, n_p, t_p, xdiff, ydiff, info_p, &
					PLAN2X, PLAN2Y, PLAN1X, PLAN1Y, &
					lowxv(0:n_p-1), upxv(0:n_p-1), lowyv(0:n_p-1), upyv(0:n_p-1), &
					UINI(1:Nx,lowyv(t_p):upyv(t_p)), &
					VINI(1:Ny,lowxv(t_p):upxv(t_p)), U2B(1:Nx,lowyv(t_p):upyv(t_p)), &
					AUINI(1:Nx,lowyv(t_p):upyv(t_p)), T1(1:ydiff*Nx),T2(1:ydiff*Nx),&
					COMP2(i), DT, EPSI, &
					Kx(lowxv(t_p):upxv(t_p)), Kxb(lowxv(t_p):upxv(t_p)), Ky(1:Ny),RHO)  
		END DO
		CALL computv(	Nx, Ny, n_p, t_p, xdiff, ydiff, info_p, &
				PLAN2X, PLAN2Y, PLAN1X, PLAN1Y, &
				lowxv(0:n_p-1), upxv(0:n_p-1), lowyv(0:n_p-1), upyv(0:n_p-1), &
				UINI(1:Nx,lowyv(t_p):upyv(t_p)), VINI(1:Ny,lowxv(t_p):upxv(t_p)), &
				T1(1:ydiff*Nx),T2(1:ydiff*Nx), &
				COMP1(2), DT, EPSI, Kx(lowxv(t_p):upxv(t_p)), Ky(1:Ny))
		time=time+DT
		! save data if taken the number of timesteps required between plots
		IF (MOD(k,plotgap).EQ.0) THEN
			CALL endwritebinfile(PLOTARRAYV(1:Ny,lowxv(t_p):upxv(t_p)),info_p,n_p,t_p,lowxv(t_p),upxv(t_p),Ny,Nx,thefilev)
			CALL endwritebinfile(PLOTARRAYU(1:Nx,lowyv(t_p):upyv(t_p)),info_p,n_p,t_p,lowyv(t_p),upyv(t_p),Nx,Ny,thefileu)
			CALL compandsave(	Nx, Ny, n_p, t_p, xdiff, ydiff, info_p, &
						PLAN1X, PLAN1Y, lowxv(0:n_p-1), upxv(0:n_p-1), lowyv(0:n_p-1), upyv(0:n_p-1), &
						UINI(1:Nx,lowyv(t_p):upyv(t_p)), VINI(1:Ny,lowxv(t_p):upxv(t_p)), &
						AUINI(1:Nx,lowyv(t_p):upyv(t_p)), &
						MA0, MASSCHECK, PLOTARRAYU(1:Nx,lowyv(t_p):upyv(t_p)), &
						PLOTARRAYV(1:Ny,lowxv(t_p):upxv(t_p)), T1(1:ydiff*Nx),T2(1:ydiff*Nx), &
						thefileu,thefilev,time,plotnum)
			IF(t_p.EQ.0) THEN
				PRINT *,'Saved data at time ',time
				PRINT *,' Mass change ',MASSCHECK
			ENDIF									
		ENDIF
	END DO 
!	fin loop
	CALL MPI_BARRIER(MPI_COMM_WORLD, info_p)
	CALL system_clock(finish,count_rate)
!	finish writing data to disk	
	CALL endwritebinfile(PLOTARRAYV(1:Ny,lowxv(t_p):upxv(t_p)),info_p,n_p,t_p,lowxv(t_p),upxv(t_p),Ny,Nx,thefilev)
	CALL endwritebinfile(PLOTARRAYU(1:Nx,lowyv(t_p):upyv(t_p)),info_p,n_p,t_p,lowyv(t_p),upyv(t_p),Nx,Ny,thefileu)

	IF(t_p.eq.0) THEN
		PRINT*,'Processor ',t_p,' took ',REAL(finish-start)/REAL(count_rate),' for execution'
		PRINT*,'There were ',n_p,' MPI processes'
		PRINT*,'The grid size was ', Nx,' by ',Ny
	END IF
	
	DEALLOCATE(	lowxv, lowyv, upxv, upyv, Kx, Kxb, Ky,  &
			PLOTARRAYV, PLOTARRAYU, UINI, U2B, VINI, AUINI, T1, T2, stat=AllocateStatus)	
	IF (AllocateStatus .NE. 0) STOP 
	IF (t_p .EQ. 0) THEN
		PRINT *,'deallocated memoire'
	ENDIF	
    
	CALL fftdestroy(PLAN1X, PLAN1Y, PLAN2X, PLAN2Y)
	CALL MPI_FINALIZE(info_p)

	END PROGRAM NLS2d
