	SUBROUTINE computu(	Nx, Ny, n_p, t_p, xdiff, ydiff, info_p, &
					 lowxv, upxv, lowyv, upyv, &
					UINI_d, AUINI_d, W, DT, EPSI, RHO )  
	!---------------------------------------------------------------------------------------------
	! .. PURPOSE ..
	!	computu solves the equation 
	!	u_t=-\rho/i*epsi*(abs(u)^2)*u 
	!	in physical space, since abs(u)^2 is time independant
	!
	! .. SPECIFICATION ..
	!
	!	computu( Nx, Ny, n_p, t_p, xdiff, ydiff, info_p, &
	!		PLAN2X, PLAN2Y, PLAN1X, PLAN1Y, lowxv_d, upxv_d, lowyv_d, upyv_d, &
	!		UINI_d, AUINI_d, W, DT, EPSI, RHO )  
	!
	!	INTEGER(kind=4)	lowxv(0:n_p-1), upxv(0:n_p-1), lowyv(0:n_p-1), upyv(0:n_p-1), &
	!			Nx, Ny, n_p, t_p, xdiff, ydiff, info_p
	!	INTEGER(kind=8)	PLAN2X, PLAN2Y, PLAN1X, PLAN1Y
	!	DOUBLE PRECISION W, DT, EPSI, RHO
	!	COMPLEX(kind=8)	AUINI_d(1:Nx, lowyv(t_p):upyv(t_p)), UINI_d(1:Nx, lowyv(t_p):upyv(t_p))
	!
	! .. INPUT ..
	! Ny		= The number of grid points in the vertical direction 
	! Nx		= The number of grid points in the horizontal direction
	! info_p	= MPI error return integer
	! n_p		= Number of MPI processes
	! t_p		= Number of this MPI process
	! lowyv_d	= lower array index for data to be saved to file
	! upyv_d	= upper array index for data to be saved to file
	! lowxv_d	= lower array index for data to be saved to file
	! upxv_d	= upper array index for data to be saved to file
	! RHO		= +1 for the defocusing case and -1 for the focusing case
	! W			= Weight for splitting
	! DT		= Time step
	! EPSI		= dispersion parameter, 0 < EPSI << 1
	! UINI_d	= u, solution of DS, size: [Nx Ny/n_p]
	! AUINI_d	= Absolute Value of u, squared; size: [Nx Ny/n_p]
	!
	! REFERENCES
	!
	! ACCURACY
	!		
	! ERROR INDICATORS AND WARNINGS
	!
	! FURTHER COMMENTS
	!---------------------------------------------------------------------------------------------
	! EXTERNAL ROUTINES REQUIRED
	! MPI library
	!---------------------------------------------------------------------------------------------

	USE mpi      
	IMPLICIT NONE    
	! .. Scalar Arguments ..
	INTEGER(KIND=4), INTENT(IN)					::  Nx, Ny, xdiff, ydiff ,n_p, t_p, info_p
	REAL(KIND=8), INTENT(IN)					::  W, DT, EPSI, RHO
	! .. Array arguments ..
	INTEGER(KIND=4),DEVICE, DIMENSION(0:n_p-1), INTENT(IN)	::  lowxv, upxv, lowyv, upyv
	COMPLEX(KIND=8),DEVICE, DIMENSION(1:Nx,lowyv(t_p):upyv(t_p)), INTENT(INOUT)	::  UINI_d
	COMPLEX(KIND=8),DEVICE, DIMENSION(1:Nx,lowyv(t_p):upyv(t_p)), INTENT(OUT)	::  AUINI_d
	! .. Local Scalars ..
	INTEGER(KIND=4)																::  i,j
	! .. Executable statements ..
	CONTINUE

	!$cuf kernel do(2) <<< *, * >>>
	DO j = t_p*ydiff+1,(t_p+1)*ydiff 
		DO i=1,Nx
			AUINI_d(i,j) = ABS(UINI_d(i,j))**2
		END DO
	END DO	
	
	!$cuf kernel do(2) <<< *, * >>>	
	DO j = t_p*ydiff+1,(t_p+1)*ydiff 
		DO i=1,Nx
			UINI_d(i,j) = UINI_d(i,j)*EXP( RHO*W*CMPLX(0d0,-1.0d0)*AUINI_d(i,j)*DT/EPSI)
		END DO
	END DO 

	END SUBROUTINE computu
