	SUBROUTINE computv(   Nx, Ny, n_p, t_p, xdiff, ydiff, info_p, &
				   PLANX, PLANY,W, DT, EPSI,  &
				   lowxv, upxv, lowyv, upyv, &
				   UINI_d, VINI_d, T1_d, T2_d, T1, T2, Kx_d, Ky_d )  
	!---------------------------------------------------------------------------------------------
	! .. PURPOSE ..
	!	computv solves the equation u_t=i*epsi*(u_xx + u_yy) in Fourier space
	!
	! .. SPECIFICATION ..
	!	computation2(	Nx, Ny, n_p, t_p, xdiff, ydiff, info_p, &
	!			PLANX, PLANY, lowxv_d, upxv_d, lowyv_d, upyv_d, &
	!			lowxv, upxv, lowyv, upyv, UINI_d, VINI_d, &
	!			T1_d, T2_d, T1, T2, W, DT, EPSI, Kx_d, Ky_d ) 
	!
	!	INTEGER(kind=4)	lowxv(0:n_p-1), upxv(0:n_p-1), lowyv(0:n_p-1), upyv(0:n_p-1), &
	!			Nx, Ny, n_p, t_p, xdiff, ydiff, info_p
	!	INTEGER(kind=8)	PLANX, PLANY
	!	REAL(kind=8)	Kx_d(lowxv(t_p):upxv(t_p)), Ky_d(1:Ny), W, DT, EPSI
	!	COMPLEX(kind=8)	UINI_d(1:Nx, lowyv(t_p):upyv(t_p)), VINI_d(1:Ny, lowxv(t_p):upxv(t_p)), &
	!			T1_d(1:ydiff*Nx), T2_d(1:ydiff*Nx), T1(1:ydiff*Nx), T2(1:ydiff*Nx)
	!
	! .. INPUT ..
	! Ny		= The number of grid points in the vertical direction 
	! Nx		= The number of grid points in the horizontal direction
	! info_p	= MPI error return integer
	! n_p		= Number of MPI processes
	! t_p		= Number of this MPI process
	! lowyv		= lower array index for data to be saved to file
	! upyv		= upper array index for data to be saved to file
	! lowxv		= lower array index for data to be saved to file
	! upxv		= upper array index for data to be saved to file
	! lowyv_d	= lower array index for data to be saved to file
	! upyv_d	= upper array index for data to be saved to file
	! lowxv_d	= lower array index for data to be saved to file
	! upxv_d	= upper array index for data to be saved to file
	! PLANX		= FFT in X direction
	! PLANY		= FFT in Y direction
	! Kx_d		= X fourier frequencies
	! Ky_d		= Y fourier frequencies
	! W			= Weight for splitting
	! DT		= Time step
	! EPSI		= dispersion parameter, 0 < EPSI << 1
	! UINI_d		= u, solution of DS, size: [Nx Ny/n_p]
	! VINI_d 		= FFT(u), (FFT in both spatial coordinates), size: [Ny Nx/n_p]
	! T1		= swap space for FFT
	! T2		= swap space for FFT
	! T1_d		= swap space for FFT
	! T2_d		= swap space for FFT
	!
	! REFERENCES
	!
	! ACCURACY
	!		
	! ERROR INDICATORS AND WARNINGS
	!
	! FURTHER COMMENTS
	!---------------------------------------------------------------------------------------------
	! EXTERNAL ROUTINES REQUIRED
	! MPI library
	! fft2dFor	-- Computes the 2d-FFT forward in both spatial coordinates, by using 
	!		MPI_ALL_TO_ALL to perform the data transposition. Data is only transposed once,
	!		so output is in the opposite orientation to input.  Input data is destroyed.
	! fft2dBack	-- Computes the 2d-FFT backward in both spatial coordinates, by using 
	!		MPI_ALL_TO_ALL to perform the data transposition. Data is only transposed once,
	!		so output is in the opposite orientation to input.  Input data is destroyed.

	USE mpi
	IMPLICIT NONE    
	! .. Scalar Arguments ..
	INTEGER(KIND=4), INTENT(IN)						::  Nx, Ny, n_p, t_p, xdiff, ydiff 
	INTEGER(KIND=4), INTENT(INOUT)					::  info_p
	INTEGER(KIND=4), INTENT(INOUT)					::  PLANX, PLANY
	REAL(KIND=8), INTENT(IN)						::  W, DT, EPSI
	! .. Array Arguments
	INTEGER(KIND=4), DIMENSION(0:n_p-1), INTENT(IN)	::  lowxv, upxv, lowyv, upyv
	REAL(KIND=8), DEVICE, DIMENSION(lowxv(t_p):upxv(t_p)), INTENT(IN)	::  Kx_d 
	REAL(KIND=8), DEVICE, DIMENSION(1:Ny), INTENT(IN)					::  Ky_d 
	COMPLEX(KIND=8), DEVICE, DIMENSION(1:Nx, lowyv(t_p):upyv(t_p)), INTENT(INOUT)	::  UINI_d
	COMPLEX(KIND=8), DEVICE, DIMENSION(1:Ny, lowxv(t_p):upxv(t_p)), INTENT(OUT)	::  VINI_d
	COMPLEX(KIND=8), DEVICE, DIMENSION(1:ydiff*Nx), INTENT(OUT)					::  T1_d, T2_d
	COMPLEX(KIND=8), DIMENSION(1:ydiff*Nx), INTENT(OUT)						::  T1, T2
	! .. Local Scalars
	INTEGER(KIND=4) 														::  i,j
	COMPLEX(KIND=8)															:: imagnum
	! .. Executable statements ..	
	CONTINUE

	CALL fft2dFor(	Nx, Ny, n_p, t_p, xdiff, ydiff, info_p, &
			PLANX, PLANY, lowxv, upxv, lowyv, upyv, &
			UINI_d, VINI_d, T1_d,T2_d,T1,T2)  
	
	!$cuf kernel do(2) <<< *, * >>>		
	DO i = t_p*xdiff+1,(t_p+1)*xdiff
		DO j=1,Ny
			imagnum=COS(W*DT*EPSI*(Kx_d(i)*Kx_d(i) + Ky_d(j)*Ky_d(j)))-&
						CMPLX(0.0,1.0)*SIN(W*DT*EPSI*(Kx_d(i)*Kx_d(i) + Ky_d(j)*Ky_d(j)))
			VINI_d(j,i)=VINI_d(j,i)*imagnum
		END DO
	END DO

	CALL fft2dBack(	Nx, Ny, n_p, t_p, xdiff, ydiff, info_p, &
			PLANX, PLANY, lowxv, upxv, lowyv, upyv, &
			VINI_d, UINI_d, T1_d,T2_d,T1,T2)  

	END SUBROUTINE computv

