	SUBROUTINE fft2dBack(	Nx, Ny, n_p, t_p, xdiff, ydiff, info_p, &
				PLANX, PLANY,  &
				lowxv, upxv, lowyv, upyv,  VARE_d, VARS_d, T1_d, T2_d, T1, T2  )  
	!---------------------------------------------------------------------------------------------
	! .. PURPOSE ..
	!	Computes the 2d-FFT backward in both spatial coordinates, by using 
	!	MPI_ALL_TO_ALL to perform the data transposition. Data is only transposed once, so
	!	output is in the opposite orientation to input. Input data is destroyed.
	!
	! .. SPECIFICATION ..
	!	fft2dBack(	Nx, Ny, n_p, t_p, xdiff, ydiff, info_p, &
	!			PLANX, PLANY, lowxv_d, upxv_d, lowyv_d, upyv_d, &
	!			lowxv, upxv, lowyv, upyv, VARE, VARS, T1_d, T2_d, T1, T2  )  
	!	INTEGER(KIND=4)	Nx, Ny, n_p, t_p, xdiff, ydiff, info_p, &
	!			lowxv(0:n_p-1), upxv(0:n_p-1), lowyv(0:n_p-1), upyv(0:n_p-1)
	!	INTEGER(KIND=8)	PLANX, PLANY
	!	COMPLEX(KIND=8)	VARE(1:Nx,lowyv(t_p):upyv(t_p)), VARS(1:Ny,lowxv(t_p):upxv(t_p)), &
	!			T1_d(1:ydiff,1:Nx), T2_d(1:ydiff,1:Nx), T1(1:ydiff,1:Nx), T2(1:ydiff,1:Nx)
	! .. INPUT ..
	! Nx		= The number of grid points in the horizontal direction
	! Ny  		= The number of grid points in the vertical direction 
	! n_p		= Number of MPI processes
	! t_p		= Number of this MPI process
	! info_p	= MPI error return integer
	! PLANX		= fft plan in X direction
	! PLANY		= fft plan in Y direction
	! lowyv		= lower array index for data to be saved to file
	! upyv		= upper array index for data to be saved to file
	! lowxv		= lower array index for data to be saved to file
	! upxv		= upper array index for data to be saved to file
	! VARE_d	= ARRAY to be transformed, size: [Nx Ny/n_p]
	! VARS_d 	= FFT(VARE), (FFT in both spatial coordinates), size: [Ny Nx/n_p]	
	! T1_d		= swap space for FFT
	! T2_d		= swap space for FFT
	! T1		= swap space for FFT
	! T2		= swap space for FFT
	!---------------------------------------------------------------------------------------------
	! EXTERNAL ROUTINES REQUIRED
	! MPI library
	! FFTW library
	!---------------------------------------------------------------------------------------------
	USE mpi
    USE precision
    USE cufft

	!IMPLICIT NONE    
	! .. Scalar Arguments ..
	INTEGER(KIND=4), INTENT(IN)						::  Nx, Ny, n_p, t_p, xdiff, ydiff
	INTEGER(KIND=4), INTENT(INOUT)					::  info_p
	INTEGER(KIND=4), INTENT(INOUT)					::  PLANX, PLANY
	! .. Array Arguments ..
	INTEGER(KIND=4), DIMENSION(0:n_p-1), INTENT(IN)	::  lowxv, upxv, lowyv, upyv
	COMPLEX(KIND=8), DEVICE, DIMENSION(1:Ny,t_p*xdiff+1:(t_p+1)*xdiff), INTENT(IN) ::  VARE_d
	COMPLEX(KIND=8), DEVICE, DIMENSION(1:Nx,t_p*ydiff+1:(t_p+1)*ydiff), INTENT(OUT) ::  VARS_d
	COMPLEX(KIND=8), DEVICE, DIMENSION(1:ydiff*Nx), INTENT(OUT)				::  T1_d, T2_d
	COMPLEX(KIND=8), DIMENSION(1:ydiff*Nx), INTENT(OUT)						::  T1, T2
	! .. Local Scalars ..
	INTEGER(KIND=4)															::  i, j, n
	! .. Executable statements ..	
	CONTINUE

	!$cuf kernel do(2) <<< *, * >>>
	DO i = t_p*xdiff+1,(t_p+1)*xdiff
		DO j=1,Ny
			T2_d(Ny*(i-(t_p*xdiff+1))+j)=VARE_d(j,i)
		END DO
	END DO	

	CALL cufftExecZ2Z(PLANY,T2_d,T1_d,CUFFT_INVERSE)
		
	!$cuf kernel do(2) <<< *, * >>>
	DO n = 1,n_p
		DO j=1,ydiff
			DO i=1,xdiff
				T2_d((j-1)*xdiff+(n-1)*xdiff*ydiff+i)=&
				T1_d(Ny*(i-1)+(n-1)*ydiff+j)/REAL(Nx*Ny,kind=8)
			END DO
		END DO
	END DO
	
	! Transfer data from GPU to CPU
	T2=T2_d
	! Perform transpose
	CALL MPI_ALLTOALL(T2(1), ydiff*xdiff, MPI_DOUBLE_COMPLEX, T1(1), ydiff*xdiff,&
              			MPI_DOUBLE_COMPLEX, MPI_COMM_WORLD, info_p)

	! Transfer data from CPU to GPU
	T1_d=T1

	!$cuf kernel do(2) <<< *, * >>>
	DO n = 1,n_p
		DO i=1,xdiff
			DO j=1,ydiff
				T2_d(Nx*(j-1)+(n-1)*xdiff+i)=&
				T1_d((j-1)*xdiff+ydiff*xdiff*(n-1)+i)
			END DO
		END DO	
	END DO

	CALL cufftExecZ2Z(PLANX,T2_d,T1_d,CUFFT_INVERSE)

	!$cuf kernel do(2) <<< *, * >>>
	DO j = t_p*ydiff+1,(t_p+1)*ydiff
		DO i=1,Nx
			VARS_d(i,j)=T1_d(Nx*(j-(t_p*ydiff+1))+i)
		END DO
	END DO
	
	END SUBROUTINE fft2dBack
