	PROGRAM NLS2d
	!-------------------------------------------------------------------------------------
	! .. PURPOSE ..
	!
	!	NLS2d solves the two dimensional cubic Schrodinger equation
	!	(i*epsi*u_t + epsi^2*u_xx + epsi^2*u_yy + 2\rho*(|u|^2)u = 0,
	!	on [-Lx*pi,Lx*pi]x[-Ly*pi,Ly*pi]
	!	by the Time Splitting Method of order 2
	!
	! .. PARAMETERS .. INITIALIZED IN INPUTFILE
	!	RHO			= +1 for the defocusing case and -1 for the focusing case
	!	Nx			= power of two, number of modes in the x direction
	!	Ny			= power of two, number of modes in the y direction
	!	Nt			= the number of timesteps
	!	plotgap		= the number of timesteps to take before plotting
	!	Lx			= definition of the periodic domain of computation in x direction
	!	Ly			= definition of the periodic domain of computation in y direction
	!	EPSI		= dispersion parameter, 0 < EPSI << 1
	!	DT			= the time step
	!
	! .. Scalars ..
	!	W1, W2, W3, W4		= coefficients of the methods, see REFERENCES
	!	t_p			= the number of the MPI process  
	!	info_p			= MPI error code
	!	n_p			= the total number of MPI processes
	!	i, k, j			= integers used for loops
	!	deb, fin, co		= variables used for the calcul of the CPU time
	!	AllocateStatus		= variable used to verify the allocation of arrays
	!	pi			= the number pi, pi=4*Arctan(1).
	!	plotnum			= number of plots saved
	!	thefileu		= handle for file to be saved to disk
	!	thefilev		= handle for file to be saved to disk
	!
	! .. Variables, 1-dimension .. 	
	!	xdiff			= Nx/n_p  Variable used for the indiciation 
	!				(and so for the partage of datas between processes) 
	!	ydiff			= Ny/n_p  Variable used for the indiciation
	!				(and so for the partage of datas between processes) 
	!	PLAN1X			= data structure containing all the information that 
	!				fftw needs in order to compute the 1D forward Fourier transform
	!				in the x direction
	!	PLAN2X			= data structure containing all the information that 
	!				fftw needs in order to compute the 1D backward Fourier transform
	!				in the x direction
	!	PLAN1Y			= data structure containing all the information that 
	!				fftw needs in order to compute the 1D forward Fourier transform
	!				in the y direction
	!	PLAN2Y			= data structure containing all the information that 
	!				fftw needs in order to compute the 1D backward Fourier transform
	!				in the y direction
	!	MASSCHECK		= test of the numerical precision provided by mass conservation
	!	MA0			= initial mass numerically calculated via FFT
	!				(in both spatial coordinates)
	!
	! .. Arrays 1d .. (Vectors)
	!	COMP1			= (/ W1, W1 /), variable used to compute 
	!				the time splitting method (see REFERENCES) in a single loop 
	!	COMP2			= (/ W2, 0 /), variable used to compute 
	!				the time splitting method (see REFERENCES) in a single loop 
	!	lowxv, upxv, lowyv, upyv = Variables used for the indication, size: n_p
	!				low=lower, up=upper, xv= x value, yv=y value
	!	Kx			= Fourier frequencies in the x direction, size: lowxv(t_p):uppxv(t_p)  
	!	Kxb			= Fourier frequencies in the x direction, modified to avoid division by zero,
	!				size: lowxv(t_p):uppxv(t_p)  
	!	Ky			= Fourier frequencies in the y direction, size: Ny	
	!
	! .. Arrays 2d .. (Matrix)
	!	Note: Sizes are indicated without parallelisation
	!
	!	UINI			= u, solution of DS, size: [Nx Ny/n_p]
	!	VINI			= FFT(u), (FFT in both spatial coordinates), size: [Ny Nx/n_p]
	!	AUINI			= Absolute Value of u, squared; size: [Nx Ny/n_p]
	!	U2B			= -((abs(u).^2) + IFFT(K.*FFT(abs(u).^2))); SIZE: [Nx Ny/n_p]
	!				where, IFFT denotes the backward FFT in both spatial coordinates,
	!				and K denotes the operator with which we solve the second equation in
	!				DS, the equation for \phi. (K=(-2*Kx.^2)./(Kx.^2 + ALPHA*Ky.^2).
	!				Also used as the array to store data to be saved to disk
	! .. REFERENCES ..
	! Yoshida, Haruo: Construction of higher order symplectic integrators, 
	!		Physics Letters A 150,(5-7) pp262-268 November 1990
	! .. ACCURACY ..
	!		
	! .. ERROR INDICATORS AND WARNINGS ..
	! One might possibly use derived datatypes, MPI_ALLTOALL_W and/or change from 
	! storing data in 2d arrays to storing data in large 1d vectors with clever indexing to 
	! reduce memory requirements
	!
	! FURTHER COMMENTS
	! This program is based on a program for the Davey Stewartson equation originally by
	! B.K. Muite and K. Roidot
	!---------------------------------------------------------------------------------------------
	! EXTERNAL ROUTINES REQUIRED
	! MPI library
	! parainit 	-- intialize parameters, which are stocked in the Input File name_file
	! computu 	-- solves the equation 
	!		u_t=-2*\rho/i*epsi*(abs(u)^2)*u 
	!		in physical space, since (abs(u)^2) is time independant
	! computv 	-- solves the equation 
	!		u_t=i*epsi*(u_xx + u_yy) in Fourier space
	! fftinit	-- initializes fftw3 plans
	! writebinfile  -- saves a two dimensional field in native binary, double format using MPIIO
	! fft2dFor  	-- Computes the 2d-FFT forward in both spatial coordinates, by using 
	!		MPI_ALL_TO_ALL to perform the data transposition. Data is only transposed once, so
	!		output is in the opposite orientation to input.  Input data is destroyed.
	! fft2dBack	-- Computes the 2d-FFT backward in both spatial coordinates, by using 
	!		MPI_ALL_TO_ALL to perform the data transposition. Data is only transposed once, so
	!		output is in the opposite orientation to input.  Input data is destroyed.
	! compandsave	-- save the Fourier coefficients and the numerical solution
	!		in a binary file, by calling the subroutine writebinfile,
	!		and calculate the test for the mass conservation.
	! fftdestroy	-- destroys fftw arrays
	! init		-- initializes the grid points of computation, fourier frequencies, 
	!		the initial data, uini, and initial mass, and
	!		also saves initial data required to calculate points.
	!-----------------------------------------------------------------------------------------------------
	USE mpi
    USE precision
    USE cufft
      
	IMPLICIT NONE    
!	Declaration of variables and parameter
	INTEGER(KIND=4)				::  Nx, Ny  
	INTEGER(KIND=4)				::  Nt, plotgap
	INTEGER(KIND=4)				::  i, k, j, start, finish, count_rate, AllocateStatus
	INTEGER(KIND=4)				::  info_p, n_p, t_p, xdiff, ydiff, plotnum, thefileu, thefilev
	INTEGER(KIND=4)				::  PLANX, PLANY
	REAL(KIND=8)				::  EPSI
	REAL(KIND=8)				::  Lx
	REAL(KIND=8)				::  Ly
	REAL(KIND=8), PARAMETER		::  pi= 3.14159265358979323846264338327950288419716939937510
	REAL(KIND=8)							::  RHO 
	REAL(KIND=8), DIMENSION(2), PARAMETER	::  COMP1=(/ 	0.5d0, 0.5d0 /)
	REAL(KIND=8), DIMENSION(2), PARAMETER	::  COMP2=(/ 1.0d0, 0.0d0  /)
	REAL(KIND=8)							::  DT, time, MASSCHECK, MA0, W  
	INTEGER, DIMENSION(MPI_STATUS_SIZE) 	::  mpistatus
	! CPU arrays
	INTEGER(KIND=4), DIMENSION(:), ALLOCATABLE	::  lowxv, upxv, lowyv, upyv
	COMPLEX(KIND=8), DIMENSION(:), ALLOCATABLE	::  T1, T2
	REAL(KIND=8), DIMENSION(:,:), ALLOCATABLE	::  PLOTARRAYU, PLOTARRAYV
	! GPU arrays
	REAL(KIND=8), DEVICE, DIMENSION(:), ALLOCATABLE			::  Kx_d, Ky_d 
	COMPLEX(KIND=8), DEVICE, DIMENSION(:), ALLOCATABLE		::  T1_d, T2_d
	COMPLEX(KIND=8), DEVICE, DIMENSION(:,:), ALLOCATABLE	::  AUINI_d, UINI_d, VINI_d
	REAL(KIND=8), DEVICE, DIMENSION(:,:), ALLOCATABLE				::  PLOTARRAYU_d, PLOTARRAYV_d
!       Initialisation of data 

	
!       Initialisation MPI environment
	CALL MPI_INIT(info_p)
	CALL MPI_COMM_SIZE(MPI_COMM_WORLD, n_p, info_p)
	CALL MPI_COMM_RANK(MPI_COMM_WORLD, t_p, info_p)   
!       give a number to my proc from 0 to n_p-1

!	Initialisation of parameter
	CALL parainit(	Nx, Ny, RHO, Nt, plotgap, Lx, Ly, EPSI, &
			DT, time, t_p, n_p, info_p )

	plotnum=0

!	indiciation of array bounds
	ydiff = Ny/n_p
	xdiff = Nx/n_p

	ALLOCATE( lowxv(0:n_p-1) , lowyv(0:n_p-1), upxv(0:n_p-1), upyv(0:n_p-1), stat=AllocateStatus)
	IF (AllocateStatus .NE. 0) STOP
	
	! Calculate array indices
	DO i = 1,n_p
		lowxv(i-1) = (i-1)*xdiff + 1
		lowyv(i-1) = (i-1)*ydiff + 1
		upxv(i-1) = i*xdiff 
		upyv(i-1) = i*ydiff
	END DO
	
!	end indiciation 

!	allocation of memory
	ALLOCATE(T1(1:ydiff*Nx),T2(1:ydiff*Nx), PLOTARRAYU(1:Nx,lowyv(t_p):upyv(t_p)),&
			 PLOTARRAYV(1:Ny,lowxv(t_p):upxv(t_p)), stat=AllocateStatus   )	 
	IF (AllocateStatus .NE. 0) STOP 

	IF (t_p.EQ.0) THEN
		PRINT *,'allocation of CPU memory, OK '
	ENDIF

	ALLOCATE(Kx_d(lowxv(t_p):upxv(t_p)),  Ky_d(1:Ny), &
			UINI_d(1:Nx,lowyv(t_p):upyv(t_p)), AUINI_d(1:Nx,lowyv(t_p):upyv(t_p)), &
			VINI_d(1:Ny,lowxv(t_p):upxv(t_p)), &
			T1_d(1:ydiff*Nx),T2_d(1:ydiff*Nx), &
			PLOTARRAYU_d(1:Nx,lowyv(t_p):upyv(t_p)),PLOTARRAYV_d(1:Ny,lowxv(t_p):upxv(t_p)),&
			 stat=AllocateStatus   )	 
	IF (AllocateStatus .NE. 0) STOP 

	IF (t_p.EQ.0) THEN
		PRINT *,'allocation of GPU memory, OK '
	ENDIF
	! use of fftw needs creation of plans: 
	CALL fftinit( Nx, Ny, xdiff, ydiff, PLANX, PLANY )
	IF (t_p.EQ.0) THEN
		PRINT *,'set up fft, OK '
	ENDIF

!	Initialisation of variables, initial data ....
	CALL init(	Nx, Ny, n_p, t_p, xdiff, ydiff, info_p, &
			PLANX, PLANY, Lx, Ly, pi, RHO, &
			MA0, lowxv, upxv, lowyv, upyv ,&
			Ky_d, Kx_d  , UINI_d , &
			AUINI_d, VINI_d, &
			T1_d,T2_d, T1, T2 )
	IF (t_p.EQ.0) THEN
	   PRINT *,'set up initial data '
	ENDIF
	
!		save initial data			
	CALL compandsave(	Nx, Ny, n_p, t_p, xdiff, ydiff, info_p, &
				PLANX, PLANY, plotnum, thefileu, thefilev, time,&
				MA0, MASSCHECK, &
				lowxv, upxv, lowyv, upyv, &
				UINI_d, VINI_d,AUINI_d, &
				PLOTARRAYU_d, PLOTARRAYV_d, PLOTARRAYU, PLOTARRAYV,&
				T1_d,T2_d,T1,T2)
!	end of the initialisation
	
	IF (t_p.EQ.0) THEN
		PRINT *,'saved initial data '
		PRINT *,' Mass change ',MASSCHECK
	ENDIF
!	system clock
	call system_clock(start,count_rate)
!	Solve the equation     
!	Start of the loop 
	DO k = 1,Nt
		DO i=1,1
			W=COMP1(i)
			CALL computv(	Nx, Ny, n_p, t_p, xdiff, ydiff, info_p, &
					PLANX, PLANY, W, DT, EPSI,  &
					lowxv, upxv, lowyv, upyv, &
					UINI_d, VINI_d, &
					T1_d,T2_d,T1, T2, &
					Kx_d, Ky_d )
			W=COMP2(i)
			CALL computu(	Nx, Ny, n_p, t_p, xdiff, ydiff, info_p, &
					lowxv, upxv, lowyv, upyv, &
					UINI_d, &
					AUINI_d,W, DT, EPSI, RHO)  
		END DO
		W=COMP1(2)
		CALL computv(	Nx, Ny, n_p, t_p, xdiff, ydiff, info_p, &
				PLANX, PLANY, W, DT, EPSI,  &
				lowxv, upxv, lowyv, upyv, &
				UINI_d, VINI_d, &
				T1_d,T2_d,T1, T2, &
				Kx_d, Ky_d )
				
		time=time+DT
		! save data if taken the number of timesteps required between plots
		IF (MOD(k,plotgap).EQ.0) THEN
			CALL endwritebinfile(PLOTARRAYV,info_p,n_p,t_p,lowxv(t_p),upxv(t_p),Ny,Nx,thefilev)
			CALL endwritebinfile(PLOTARRAYU,info_p,n_p,t_p,lowyv(t_p),upyv(t_p),Nx,Ny,thefileu)
			CALL compandsave(	Nx, Ny, n_p, t_p, xdiff, ydiff, info_p, &
				PLANX, PLANY, plotnum, thefileu, thefilev, time,&
				MA0, MASSCHECK, &
				lowxv, upxv, lowyv, upyv, &
				UINI_d, VINI_d,AUINI_d, &
				PLOTARRAYU_d, PLOTARRAYV_d, PLOTARRAYU, PLOTARRAYV,&
				T1_d,T2_d,T1,T2)
			IF(t_p.EQ.0) THEN
				PRINT *,' Saved data at time ',time
				PRINT *,' Mass change ',MASSCHECK
			ENDIF									
		ENDIF
	END DO 
!	fin loop
	CALL MPI_BARRIER(MPI_COMM_WORLD, info_p)
	CALL system_clock(finish,count_rate)
!	finish writing data to disk	
	CALL endwritebinfile(PLOTARRAYV,info_p,n_p,t_p,lowxv(t_p),upxv(t_p),Ny,Nx,thefilev)
	CALL endwritebinfile(PLOTARRAYU,info_p,n_p,t_p,lowyv(t_p),upyv(t_p),Nx,Ny,thefileu)

	IF(t_p.eq.0) THEN
		PRINT*,'Processor ',t_p,' took ',REAL(finish-start)/REAL(count_rate),' for execution'
		PRINT*,'There were ',n_p,' MPI processes'
		PRINT*,'The grid size was ', Nx,' by ',Ny
	END IF
	
	DEALLOCATE(lowxv, upxv, lowyv, upyv, T1, T2, PLOTARRAYU, PLOTARRAYV, stat=AllocateStatus)	 
	IF (AllocateStatus .NE. 0) STOP 

	IF (t_p.EQ.0) THEN
		PRINT *,'Deallocation of CPU memory, OK '
	ENDIF

	DEALLOCATE(Kx_d,  Ky_d, UINI_d, AUINI_d, &
			VINI_d, T1_d,T2_d, PLOTARRAYU_d,PLOTARRAYV_d, stat=AllocateStatus   )	 
	IF (AllocateStatus .NE. 0) STOP 

	IF (t_p.EQ.0) THEN
		PRINT *,'Deallocation of GPU memory, OK '
	ENDIF

	CALL fftdestroy(PLANX, PLANY)
	CALL MPI_FINALIZE(info_p)

	END PROGRAM NLS2d
