                                                                     
                                                                     
                                                                     
                                             
 	!--------------------------------------------------------------------
	!
	! PURPOSE
	!
	! This program numerically solves the 2D incompressible Navier-Stokes
	! on a Square Domain [0,1]x[0,1] using pseudo-spectral methods and
	! Crank-Nicolson timestepping. The numerical solution is compared to
	! the exact Taylor-Green Vortex Solution. 
	!
	! AUTHORS
	!
	! B. Cloutier, B.K. Muite, P. Rigge
	! 4 June 2012
	!
	! Periodic free-slip boundary conditions and Initial conditions:
	!	u(x,y,0)=sin(2*pi*x)cos(2*pi*y)
	!	v(x,y,0)=-cos(2*pi*x)sin(2*pi*y)
	! Analytical Solution (subscript denote derivatives):
	!	u(x,y,t)=sin(2*pi*x)cos(2*pi*y)exp(-8*pi^2*t/Re)
	!	v(x,y,t)=-cos(2*pi*x)sin(2*pi*y)exp(-8*pi^2*t/Re)
	!   u_y(x,y,t)=-2*pi*sin(2*pi*x)sin(2*pi*y)exp(-8*pi^2*t/Re)
	!	v_x(x,y,t)=2*pi*sin(2*pi*x)sin(2*pi*y)exp(-8*pi^2*t/Re)
	!	omega=v_x-u_y
	!
	! .. Parameters ..
	!  Nx				= number of modes in x - power of 2 for FFT
	!  Ny				= number of modes in y - power of 2 for FFT
	!  nplots			= number of plots produced
	!  plotgap			= number of timesteps inbetween plots
	!  Re 				= dimensionless Reynold's number
	!  ReInv			= 1/Re for optimization
	!  dt				= timestep size 
	!  dtInv			= 1/dt for optimization
	!  tol				= determines when convergences is reached
	!  scalemodes		= 1/(Nx*Ny), scaling after preforming FFTs
	!  numthreads		= number of CPUs used in calculation
	! .. Scalars ..
	!  i				= loop counter in x direction
	!  j				= loop counter in y direction
	!  n				= loop counter for timesteps direction	
	!  allocatestatus	= error indicator during allocation
	!  time				= times at which data is saved
	!  chg				= error at each iteration	
	! .. Arrays ..
	!  omeg				= vorticity	in real space
	!  omeghat			= 2D Fourier transform of vorticity
	!						at next iterate
	!  omeghat_temp		= used so omeghat isn't lost when preforming
	!						c2r and can be used later in calculation.
	!						See timestepping scheme for more.
	!  omegoldhat		= 2D Fourier transform of vorticity at previous
	!						iterate
	!  nl				= nonlinear term 
	!  nlhat			= nonlinear term in Fourier space
	!  nloldhat			= nonlinear term in Fourier space
	!						at previous iterate
	!  omegexact		= taylor-green vorticity at
	!						at final step
	!  psihat			= 2D Fourier transform of streamfunction
	!						at next iteration
	!  temp1/temp2/temp3= reusable complex/real space used for
	!						calculations. This reduces number of
	!						arrays stored.
	! .. Vectors ..
	!  kx				= fourier frequencies in x direction
	!  ky				= fourier frequencies in y direction
	!  x				= x locations
	!  y				= y locations
	!  name_config		= array to store filename for data to be saved    		
	! REFERENCES
	!
	! ACKNOWLEDGEMENTS
	!
	! ACCURACY
	!		
	! ERROR INDICATORS AND WARNINGS
	!
	! FURTHER COMMENTS
	!--------------------------------------------------------------------	

	PROGRAM main	
	USE omp_lib
	IMPLICIT NONE
	! declare variables		
    INTEGER(kind=4), PARAMETER 	::  Nx=64
	INTEGER(kind=4), PARAMETER 	::  Ny=64
	REAL(kind=8), PARAMETER		::  dt=0.000125d0
	INTEGER(kind=4), PARAMETER	::	nplots=1, plotgap=20
	REAL(kind=8), PARAMETER	&
		::  pi=3.14159265358979323846264338327950288419716939937510d0
	REAL(kind=8), PARAMETER		::  Re=1.0d0 		
	REAL(kind=8), PARAMETER		::	ReInv=1.0d0/REAL(Re,kind(0d0)),&
									dtInv=1.0d0/REAL(dt,kind(0d0)) 
	REAL(kind=8), PARAMETER		::	tol=0.1d0**1
	REAL(kind=8)				::  scalemodes,chg
	COMPLEX(kind=8), DIMENSION(:), ALLOCATABLE		::  kx,ky						
	REAL(kind=8), DIMENSION(:), ALLOCATABLE			::  x,y,time
	REAL(kind=8), DIMENSION(:,:), ALLOCATABLE		::	omeg,nl,temp2,temp3,omegexact
	COMPLEX(kind=8), DIMENSION(:,:), ALLOCATABLE	:: omegoldhat, nloldhat,&
													   omeghat,nlhat, psihat,&
													   omeghat_temp,temp1
	INTEGER(kind=4)								::  i,j,n,t,allocatestatus
	!FFTW variables
	INTEGER(kind=4)								::  ierr
	INTEGER(kind=4), PARAMETER      			::  FFTW_IN_PLACE = 8,&
													FFTW_MEASURE = 0,&
										    		FFTW_EXHAUSTIVE = 8,&
													FFTW_PATIENT = 32,&
                    	 				    		FFTW_ESTIMATE = 64
    INTEGER(kind=4),PARAMETER       			::  FFTW_FORWARD = -1,&
													FFTW_BACKWARD=1	
	INTEGER(kind=8)								::  planfxy,planbxy
	!variables used for saving data and timing
	INTEGER(kind=4)								::  count, iol 
	CHARACTER*100			 					::  name_config
	INTEGER(kind=4)								::	start, finish, count_rate 
	
    PRINT *,'Grid:',Nx,'X',Ny
	PRINT *,'dt:',dt	
	ALLOCATE(time(1:nplots+1),kx(1:Nx),ky(1:Ny),x(1:Nx),y(1:Ny),&
			omeg(1:Nx,1:Ny),omegoldhat(1:Nx/2+1,1:Ny),nloldhat(1:Nx/2+1,1:Ny),&
			omeghat(1:Nx/2+1,1:Ny),nl(1:Nx,1:Ny),nlhat(1:Nx/2+1,1:Ny),&
			psihat(1:Nx/2+1,1:Ny),omegexact(1:Nx,1:Ny),omeghat_temp(1:Nx/2+1,Ny),&
			temp1(1:Nx/2+1,1:Ny),temp2(1:Nx,1:Ny),temp3(1:Nx,1:Ny),stat=AllocateStatus)	
	IF (AllocateStatus .ne. 0) STOP 
	PRINT *,'allocated space'
	
	! set up multithreaded ffts
	CALL dfftw_plan_dft_r2c_2d_(planfxy,Nx,Ny,omeg,omeghat,FFTW_ESTIMATE)
	CALL dfftw_plan_dft_c2r_2d_(planbxy,Nx,Ny,omeghat,omeg,FFTW_ESTIMATE)
	PRINT *,'Setup 2D FFTs'

	! setup fourier frequencies in x-direction
	DO i=1,Nx/2+1
		kx(i)= 2.0d0*pi*cmplx(0.0d0,1.0d0)*REAL(i-1,kind(0d0))  			
	END DO
	kx(1+Nx/2)=0.0d0
	DO i = 1,Nx/2 -1
		kx(i+1+Nx/2)=-kx(1-i+Nx/2)
	END DO	
	DO i=1,Nx
		x(i)=REAL(i-1,kind(0d0))/REAL(Nx,kind(0d0)) 
	END DO
	! setup fourier frequencies in y-direction
	DO j=1,Ny/2+1
		ky(j)= 2.0d0*pi*cmplx(0.0d0,1.0d0)*REAL(j-1,kind(0d0))  			
	END DO	
	ky(1+Ny/2)=0.0d0
	DO j = 1,Ny/2 -1
		ky(j+1+Ny/2)=-ky(1-j+Ny/2)
	END DO	
	DO j=1,Ny
		y(j)=REAL(j-1,kind(0d0))/REAL(Ny,kind(0d0)) 
	END DO	
	scalemodes=1.0d0/REAL(Nx*Ny,kind(0d0))
	PRINT *,'Setup grid and fourier frequencies'

	!initial data
	DO j=1,Ny
		DO i=1,NX
			omeg(i,j)=4.0d0*pi*sin(2.0d0*pi*x(i))*sin(2.0d0*pi*y(j))
		END DO
	END DO	
	!\hat{\omega^{n,k}}
	CALL dfftw_execute_dft_r2c_(planfxy,omeg,omeghat)
	!save to use later, or else it will be lost when preforming c2r fft
	DO j=1,Ny
		DO i=1,Nx/2+1
			omeghat_temp(i,j)=omeghat(i,j)
		END DO
	END DO 
	
	!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
	!get initial nonlinear term using omeghat to find psihat, u, and v!
	!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
	DO j=1,Ny
		DO i=1,Nx/2+1
			psihat(i,j)=-omeghat(i,j)/(kx(i)*kx(i)+ky(j)*ky(j)+0.10d0**14)
		END DO
	END DO
		
	DO j=1,Ny
		DO i=1,Nx/2+1
			temp1(i,j)=psihat(i,j)*ky(j)*scalemodes
		END DO
	END DO			
	CALL dfftw_execute_dft_c2r_(planbxy,temp1,temp3) !u
	
	! \hat{\omega_x^{n,k}}
	DO j=1,Ny
		DO i=1,Nx/2+1
			temp1(i,j)=omeghat(i,j)*kx(i)
		END DO
	END DO
	CALL dfftw_execute_dft_c2r_(planbxy,temp1,temp2) ! \omega_x^{n,k}
	! first part nonlinear term in real space
	DO j=1,Ny
		DO i=1,Nx
			nl(i,j)=temp3(i,j)*temp2(i,j)
		END DO
	END DO

	DO j=1,Ny
		DO i=1,Nx/2+1
			temp1(i,j)=-psihat(i,j)*kx(i)*scalemodes
		END DO
	END DO
	CALL dfftw_execute_dft_c2r_(planbxy,temp1,temp3) !v	
	! \hat{\omega_y^{n,k}}
	DO j=1,Ny
		DO i=1,Nx/2+1
			temp1(i,j)=omeghat(i,j)*ky(j)
		END DO
	END DO
	CALL dfftw_execute_dft_c2r_(planbxy,temp1,temp2) ! \omega_y^{n,k}
	! get the rest of nonlinear term in real space
	DO j=1,Ny
		DO i=1,Nx
			nl(i,j)=(nl(i,j)+temp3(i,j)*temp2(i,j))*scalemodes
		END DO
	END DO
	CALL dfftw_execute_dft_r2c_(planfxy,nl,nlhat) 
	!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
	
	CALL dfftw_execute_dft_c2r_(planbxy,omeghat_temp,omeg)
	DO j=1,Ny
		DO i=1,Nx				
			temp2(i,j)=omeg(i,j)	
		END DO
	END DO 
	
	PRINT *,'Got initial data, starting timestepping'	
	time(1)=0.0d0
	CALL system_clock(start,count_rate)
	DO t=1,nplots
		DO n=1,plotgap
			chg=1.0d0
			! save old values(__^{n,k} terms in equation)
			DO j=1,Ny
				DO i=1,Nx/2+1
					nloldhat(i,j)=nlhat(i,j)
				END DO
			END DO
			DO j=1,Ny
				DO i=1,Nx/2+1
					omegoldhat(i,j)=omeghat(i,j)
				END DO
			END DO
			DO WHILE (chg>tol)
				!Crank-Nicolson timestepping to get \hat{\omega^{n+1,k+1}}
				DO j=1,Ny
					DO i=1,Nx/2+1
						omeghat(i,j)=( (dtInv+0.5d0*ReInv*(kx(i)*kx(i)+ky(j)*ky(j)))&
								*omegoldhat(i,j) - 0.5d0*(nloldhat(i,j)+nlhat(i,j)))/&
								(dtInv-0.5d0*ReInv*(kx(i)*kx(i)+ky(j)*ky(j)))  
					END DO
				END DO 
				DO j=1,Ny
					DO i=1,Nx/2+1
						omeghat_temp(i,j)=omeghat(i,j)
					END DO
				END DO 
				CALL dfftw_execute_dft_c2r_(planbxy,omeghat_temp,omeg)
				! check for convergence
				chg=0.0d0
				DO j=1,Ny
					DO i=1,Nx
						chg=chg+(omeg(i,j)-temp2(i,j))*(omeg(i,j)-temp2(i,j))&
						*scalemodes*scalemodes
					END DO
				END DO
				!!!!!!!!!!!!!!!!!!!!!!!!!!!!
				!Calculate _{n+1,k+1} terms!
				!!!!!!!!!!!!!!!!!!!!!!!!!!!!
				DO j=1,Ny
					DO i=1,Nx/2+1
						psihat(i,j)=-omeghat(i,j)/(kx(i)*kx(i)+ky(j)*ky(j)+0.10d0**14)
					END DO
				END DO
				
				DO j=1,Ny
					DO i=1,Nx/2+1
						temp1(i,j)=psihat(i,j)*ky(j)*scalemodes
					END DO
				END DO			
				CALL dfftw_execute_dft_c2r_(planbxy,temp1,temp3) !u

				! \hat{\omega_x^{n,k}}
				DO j=1,Ny
					DO i=1,Nx/2+1
						temp1(i,j)=omeghat(i,j)*kx(i)
					END DO
				END DO
				CALL dfftw_execute_dft_c2r_(planbxy,temp1,temp2) ! \omega_x^{n,k}
				! first part nonlinear term in real space
				DO j=1,Ny
					DO i=1,Nx
						nl(i,j)=temp3(i,j)*temp2(i,j)
					END DO
				END DO

				DO j=1,Ny
					DO i=1,Nx/2+1
						temp1(i,j)=-psihat(i,j)*kx(i)*scalemodes
					END DO
				END DO
				CALL dfftw_execute_dft_c2r_(planbxy,temp1,temp3) !v	
				! \hat{\omega_y^{n,k}}
				DO j=1,Ny
					DO i=1,Nx/2+1
						temp1(i,j)=omeghat(i,j)*ky(j)
					END DO
				END DO
				CALL dfftw_execute_dft_c2r_(planbxy,temp1,temp2) ! \omega_y^{n,k}
				! get the rest of nonlinear term in real space
				DO j=1,Ny
					DO i=1,Nx
						nl(i,j)=(nl(i,j)+temp3(i,j)*temp2(i,j))*scalemodes
					END DO
				END DO
				CALL dfftw_execute_dft_r2c_(planfxy,nl,nlhat) 
				!!!!!!!!!!!!!!!!!!!!!!!!!!!!
				
				!\omega^{n+1,k+1} is saved for next iteration
				DO j=1,Ny
					DO i=1,Nx
						temp2(i,j)=omeg(i,j)
					END DO
				END DO
			END DO
		END DO	
		time(t+1)=time(t)+dt*plotgap
		!PRINT *, time(t+1)
	END DO
	CALL system_clock(finish,count_rate)
	PRINT*,'Program took ',REAL(finish-start)/REAL(count_rate),&
			'for Time stepping'
		
	!get exact omega
	DO j=1,Ny
		DO i=1,Nx
			omegexact(i,j)=4.0d0*pi*sin(2.0d0*pi*x(i))*&
				sin(2.0d0*pi*y(j))*exp(-8.0d0*ReInv*pi**2*nplots*plotgap*dt)
		END DO
	END DO
	
	!compute max error
	PRINT *,'Max Error:',maxval(abs(omeg*scalemodes-omegexact))		
	
	!!!!!!!!!!!!!!!!!!!!!!!!
	!copy over data to disk!
	!!!!!!!!!!!!!!!!!!!!!!!!
	write(name_config,'(a,i0,a)') 'omega',1,'.datbin'
	INQUIRE(iolength=iol) omeg(1,1)
	OPEN(unit=11,FILE=name_config,form="unformatted", access="direct",recl=iol) 
	count = 1	
	DO j=1,Ny
		DO i=1,Nx
			WRITE(11,rec=count) omeg(i,j)*scalemodes
			count=count+1
		END DO
	END DO
	CLOSE(11)
			
	name_config = 'xcoord.dat' 
	OPEN(unit=11,FILE=name_config,status="UNKNOWN") 	
	REWIND(11)
	DO i=1,Nx
		WRITE(11,*) x(i)
	END DO
	CLOSE(11)

	name_config = 'ycoord.dat' 
	OPEN(unit=11,FILE=name_config,status="UNKNOWN") 	
	REWIND(11)
	DO j=1,Ny
		WRITE(11,*) y(j)
	END DO
	CLOSE(11)
	!!!!!!!!!!!!!!!!!!!!!!!!

	CALL dfftw_destroy_plan_(planfxy)
	CALL dfftw_destroy_plan_(planbxy)	

	DEALLOCATE(time,temp1,temp2,temp3,kx,ky,x,y,&
			omeg,omegoldhat,omegexact, nloldhat,&
			omeghat,nl, nlhat, psihat, omeghat_temp,&
			stat=AllocateStatus)	
	IF (AllocateStatus .ne. 0) STOP
	PRINT *,'Program execution complete'
	END PROGRAM main