	SUBROUTINE fft2dBack(	Nx, Ny, n_p, t_p, xdiff, ydiff, info_p, &
				PLAN2X, PLAN2Y, lowxv, upxv, lowyv, upyv, &
				VARE, VARS, T1, T2, VARET, VARST  )  
	!---------------------------------------------------------------------------------------------
	! .. PURPOSE ..
	!	Computes the 2d-FFT backward in both spatial coordinates, by using 
	!	MPI_ALL_TO_ALL to perform the data transposition. Data is only transposed once, so
	!	output is in the opposite orientation to input. Input data is destroyed.
	!
	! .. SPECIFICATION ..
	!	fft2dBack(	Nx, Ny, n_p, t_p, xdiff, ydiff, info_p, &
	!			PLAN2X, PLAN2Y, lowxv, upxv, lowyv, upyv, &
	!			VARE, VARS, T1, T2  )  
	!	INTEGER(KIND=4)	Nx, Ny, n_p, t_p, xdiff, ydiff, info_p, &
	!			lowxv(0:n_p-1), upxv(0:n_p-1), lowyv(0:n_p-1), upyv(0:n_p-1)
	!	INTEGER(KIND=8)	PLAN2X, PLAN2Y
	!	COMPLEX(KIND=8)	VARE(1:Nx,lowyv(t_p):upyv(t_p)), VARS(1:Ny,lowxv(t_p):upxv(t_p)), &
	!			T1((1:ydiff,1:Nx)), T2((1:ydiff,1:Nx))
	! .. INPUT ..
	! Nx		= The number of grid points in the horizontal direction
	! Ny  		= The number of grid points in the vertical direction 
	! n_p		= Number of MPI processes
	! t_p		= Number of this MPI process
	! info_p	= MPI error return integer
	! PLAN2X	= Backward fft in X direction
	! PLAN2Y	= Backward fft in Y direction
	! lowyv		= lower array index for data to be saved to file
	! upyv		= upper array index for data to be saved to file
	! lowxv		= lower array index for data to be saved to file
	! upxv		= upper array index for data to be saved to file
	! VARE		= ARRAY to be transformed, size: [Nx Ny/n_p]
	! VARS 		= FFT(VARE), (FFT in both spatial coordinates), size: [Ny Nx/n_p]	
	! T1		= swap space for FFT
	! T2		= swap space for FFT
	!---------------------------------------------------------------------------------------------
	! EXTERNAL ROUTINES REQUIRED
	! MPI library
	! FFTW library
	!---------------------------------------------------------------------------------------------
	USE mpi
	IMPLICIT NONE    
	! .. Scalar Arguments ..
	INTEGER(KIND=4), INTENT(IN)						::  Nx, Ny, n_p, t_p, xdiff, ydiff
	INTEGER(KIND=4), INTENT(INOUT)					::  info_p
	INTEGER(KIND=8), INTENT(INOUT)					::  PLAN2X, PLAN2Y
	! .. Array Arguments ..
	INTEGER(KIND=4), DIMENSION(0:n_p-1), INTENT(IN)	::  lowxv, upxv, lowyv, upyv
	COMPLEX(KIND=8), DIMENSION(1:Ny,lowxv(t_p):upxv(t_p)), INTENT(INOUT) 	::  VARE
	COMPLEX(KIND=8), DIMENSION(1:Nx,lowyv(t_p):upyv(t_p)), INTENT(OUT) 		::  VARS
	COMPLEX(KIND=8), DIMENSION(1:ydiff*Nx), INTENT(OUT)						::  T1, T2
	COMPLEX(KIND=8), DIMENSION(1:Ny,lowxv(t_p):upxv(t_p)), INTENT(OUT) 		::  VARET
	COMPLEX(KIND=8), DIMENSION(1:Nx,lowyv(t_p):upyv(t_p)), INTENT(OUT) 		::  VARST
	! .. Local Scalars ..
	INTEGER(KIND=4)															::  i, j, n
	! .. Executable statements ..	
	CONTINUE

	DO i = lowxv(t_p),upxv(t_p)
		CALL dfftw_execute_dft(PLAN2Y,VARE(1:Ny,i),VARET(1:Ny,i))
	END DO 

	DO n = 1,n_p
		DO j=1,ydiff
			DO i=1,xdiff
				T1(1+(i-1)*ydiff+(n-1)*xdiff*ydiff+j-1)=&
				VARET(lowyv(n-1)+j-1, lowxv(t_p)+i-1)/REAL(Nx*Ny,kind(0d0))
			END DO
		END DO
	END DO
	CALL MPI_ALLTOALL(T1(1), ydiff*xdiff, MPI_DOUBLE_COMPLEX, T2(1), ydiff*xdiff,&
              			MPI_DOUBLE_COMPLEX, MPI_COMM_WORLD, info_p)
	DO n = 1,n_p
		DO i=1,xdiff
			DO j=1,ydiff
				VARST(lowxv(n-1)+i-1, lowyv(t_p)+j-1)=&
				T2(1+(i-1)*ydiff+ydiff*xdiff*(n-1)+j-1)
			END DO
		END DO	
	END DO

	DO j=lowyv(t_p),upyv(t_p)
		CALL dfftw_execute_dft(PLAN2X,VARST(1:Nx,j),VARS(1:Nx,j))
	END DO

	END SUBROUTINE fft2dBack
