	PROGRAM NS2d
	!-------------------------------------------------------------------------------------
	! .. PURPOSE ..
	!
	!	NS2d solves the two dimensional Navier Stokes equations
	! 	on a Square Domain [0,1]x[0,1] using pseudo-spectral methods and
	! 	Crank-Nicolson timestepping. The numerical solution is compared to
	! 	the exact Taylor-Green Vortex Solution. 
	!
	! .. PARAMETERS .. INITIALIZED IN INPUTFILE
	!	Re			= Reynolds number
	!	Nx			= power of two, number of modes in the x direction
	!	Ny			= power of two, number of modes in the y direction
	!	Nt			= the number of timesteps
	!	plotgap		= the number of timesteps to take before plotting
	!	Lx			= definition of the periodic domain of computation in x direction
	!	Ly			= definition of the periodic domain of computation in y direction
	!	EPSI		= dispersion parameter, 0 < EPSI << 1
	!	DT			= the time step
	!
	! .. Scalars ..
	!	W1, W2, W3, W4		= coefficients of the methods, see REFERENCES
	!	t_p			= the number of the MPI process  
	!	info_p			= MPI error code
	!	n_p			= the total number of MPI processes
	!	i, k, j			= integers used for loops
	!	deb, fin, co		= variables used for the calcul of the CPU time
	!	AllocateStatus		= variable used to verify the allocation of arrays
	!	pi			= the number pi, pi=4*Arctan(1).
	!	plotnum			= number of plots saved
	!	thefileu		= handle for file to be saved to disk
	!	thefilev		= handle for file to be saved to disk
	!
	! .. Variables, 1-dimension .. 	
	!	xdiff		= Nx/n_p  Variable used for the indiciation 
	!				(and so for the partage of datas between processes) 
	!	ydiff		= Ny/n_p  Variable used for the indiciation
	!				(and so for the partage of datas between processes) 
	!	PLAN1X		= data structure containing all the information that 
	!				fftw needs in order to compute the 1D forward Fourier transform
	!				in the x direction
	!	PLAN2X		= data structure containing all the information that 
	!				fftw needs in order to compute the 1D backward Fourier transform
	!				in the x direction
	!	PLAN1Y		= data structure containing all the information that 
	!				fftw needs in order to compute the 1D forward Fourier transform
	!				in the y direction
	!	PLAN2Y		= data structure containing all the information that 
	!				fftw needs in order to compute the 1D backward Fourier transform
	!				in the y direction
	!	MASSCHECK	= test of the numerical precision provided by mass conservation
	!	MA0			= initial mass numerically calculated via FFT
	!				(in both spatial coordinates)
	!
	! .. Arrays 1d .. (Vectors)
	!	COMP1		= (/ W1, W1 /), variable used to compute 
	!				the time splitting method (see REFERENCES) in a single loop 
	!	COMP2		= (/ W2, 0 /), variable used to compute 
	!				the time splitting method (see REFERENCES) in a single loop 
	!	lowxv, upxv, lowyv, upyv = Variables used for the indication, size: n_p
	!				low=lower, up=upper, xv= x value, yv=y value
	!	Kx			= Fourier frequencies in the x direction, size: lowxv(t_p):uppxv(t_p)  
	!	Kxb			= Fourier frequencies in the x direction, modified to avoid division by zero,
	!				size: lowxv(t_p):uppxv(t_p)  
	!	Ky			= Fourier frequencies in the y direction, size: Ny	
	!
	! .. Arrays 2d .. (Matrix)
	!	Note: Sizes are indicated without parallelisation
	!
	!	UINI		= u, solution of DS, size: [Nx Ny/n_p]
	!	VINI		= FFT(u), (FFT in both spatial coordinates), size: [Ny Nx/n_p]
	!	AUINI		= Absolute Value of u, squared; size: [Nx Ny/n_p]
	!	U2B			= -((abs(u).^2) + IFFT(K.*FFT(abs(u).^2))); SIZE: [Nx Ny/n_p]
	!				where, IFFT denotes the backward FFT in both spatial coordinates,
	!				and K denotes the operator with which we solve the second equation in
	!				DS, the equation for \phi. (K=(-2*Kx.^2)./(Kx.^2 + ALPHA*Ky.^2).
	!				Also used as the array to store data to be saved to disk
	! .. REFERENCES ..
	!
	! .. ACCURACY ..
	!		
	! .. ERROR INDICATORS AND WARNINGS ..
	! One might possibly use derived datatypes, MPI_ALLTOALL_W and/or change from 
	! storing data in 2d arrays to storing data in large 1d vectors with clever indexing to 
	! reduce memory requirements
	!
	! FURTHER COMMENTS
	! This program is based on a program for the Davey Stewartson equation originally by
	! B.K. Muite and K. Roidot
	!---------------------------------------------------------------------------------------------
	! EXTERNAL ROUTINES REQUIRED
	! MPI library
	! parainit 	-- intialize parameters, which are stored in the Input File 
	! fftinit	-- initializes fftw3 plans
	! writebinfile  -- saves a two dimensional field in native binary, double format using MPIIO
	! fft2dFor  	-- Computes the 2d-FFT forward in both spatial coordinates, by using 
	!		MPI_ALL_TO_ALL to perform the data transposition. Data is only transposed once, so
	!		output is in the opposite orientation to input.  Input data is destroyed.
	! fft2dBack	-- Computes the 2d-FFT backward in both spatial coordinates, by using 
	!		MPI_ALL_TO_ALL to perform the data transposition. Data is only transposed once, so
	!		output is in the opposite orientation to input.  Input data is destroyed.
	! compandsave	-- save the Fourier coefficients and the numerical solution
	!		in a binary file, by calling the subroutine writebinfile,
	!		and calculate the test for the mass conservation.
	! fftdestroy	-- destroys fftw arrays
	! init		-- initializes the grid points of computation, fourier frequencies, 
	!		the initial data, uini, and initial mass, and
	!		also saves initial data required to calculate points.
	!-----------------------------------------------------------------------------------------------------
	USE mpi
      
	IMPLICIT NONE    
!	Declaration of variables and parameter
	INTEGER(KIND=4)				::  Nx=2048, Ny=2048  
	INTEGER(KIND=4)				::  Nt=20, plotgap=20
	INTEGER(KIND=4)				::  i, k, j, start, finish, count_rate, AllocateStatus
	INTEGER(KIND=4)				::  info_p, n_p, t_p, xdiff, ydiff, plotnum, thefileu, thefilev
	REAL(KIND=8)				::  Re=1.0d0		
	REAL(KIND=8)				::	ReInv
	REAL(KIND=8), PARAMETER		::	tol=0.1d0**10
	REAL(KIND=8)				::  chg, mychg
	REAL(KIND=8)				::  Lx, time
	REAL(KIND=8)				::  Ly, x, y
	REAL(KIND=8), PARAMETER		::  pi= 3.14159265358979323846264338327950288419716939937510
	REAL(KIND=8)				::  dt=0.000125d0
	REAL(KIND=8)				::  dtInv
	REAL(KIND=8)				::  calcerr   
	INTEGER(KIND=8)				::  PLAN1X, PLAN1Y, PLAN2X, PLAN2Y
	
	INTEGER, DIMENSION(MPI_STATUS_SIZE) 			::  mpistatus
	! CPU arrays
	INTEGER(KIND=4), DIMENSION(:), ALLOCATABLE		::  lowxv, upxv, lowyv, upyv
	COMPLEX(KIND=8), DIMENSION(:), ALLOCATABLE		::  T1, T2
	REAL(KIND=8), DIMENSION(:,:), ALLOCATABLE		::  PLOTARRAYU, PLOTARRAYV
	REAL(KIND=8), DIMENSION(:), ALLOCATABLE			::  Kx, Ky, Ky2, Kx2 
	COMPLEX(KIND=8), DIMENSION(:,:), ALLOCATABLE	::  omeg, omeghat, nl, &
															rhsfixhat, ffttemp,ffttemphat, &
															nlhat, temp1, temp2, temp3
!       Initialisation MPI environment
	CALL MPI_INIT(info_p)
	CALL MPI_COMM_SIZE(MPI_COMM_WORLD, n_p, info_p)
	CALL MPI_COMM_RANK(MPI_COMM_WORLD, t_p, info_p)   
!       give a number to my proc from 0 to n_p-1

	plotnum=0
	CALL parainit(	Nx, Ny, Re, Nt, plotgap, &
			DT, time, t_p, n_p, info_p )
	dtInv=1.0d0/DT
	ReInv=1.0d0/Re
!	indiciation of array bounds
	ydiff = Ny/n_p
	xdiff = Nx/n_p

	ALLOCATE( lowxv(0:n_p-1) , lowyv(0:n_p-1), upxv(0:n_p-1), upyv(0:n_p-1), stat=AllocateStatus)
	IF (AllocateStatus .NE. 0) STOP
	
	! Calculate array indices
	DO i = 1,n_p
		lowxv(i-1) = (i-1)*xdiff + 1
		lowyv(i-1) = (i-1)*ydiff + 1
		upxv(i-1) = i*xdiff 
		upyv(i-1) = i*ydiff
	END DO
	
!	end indiciation 

!	allocation of memory

	ALLOCATE(Kx(t_p*xdiff+1:(t_p+1)*xdiff),Ky(1:Ny),Kx2(t_p*xdiff+1:(t_p+1)*xdiff),Ky2(1:Ny),&
			omeg(1:Nx,t_p*ydiff+1:(t_p+1)*ydiff), nl(1:Nx,t_p*ydiff+1:(t_p+1)*ydiff), &
			temp2(1:Nx,t_p*ydiff+1:(t_p+1)*ydiff), temp3(1:Nx,t_p*ydiff+1:(t_p+1)*ydiff), &
			omeghat(1:Ny,t_p*xdiff+1:(t_p+1)*xdiff), nlhat(1:Ny,t_p*xdiff+1:(t_p+1)*xdiff), & 
			rhsfixhat(1:Ny,t_p*xdiff+1:(t_p+1)*xdiff),&
			ffttemp(1:Nx,t_p*ydiff+1:(t_p+1)*ydiff), ffttemphat(1:Ny,t_p*xdiff+1:(t_p+1)*xdiff), &
			temp1(1:Ny,t_p*xdiff+1:(t_p+1)*xdiff), T1(1:ydiff*Nx), T2(1:ydiff*Nx), &
			PLOTARRAYU(1:Nx,t_p*ydiff+1:(t_p+1)*ydiff),PLOTARRAYV(1:Ny,t_p*xdiff+1:(t_p+1)*xdiff),&
			 stat=AllocateStatus   )	 
	IF (AllocateStatus .NE. 0) STOP 

	IF (t_p.EQ.0) THEN
		PRINT *,'allocation of CPU memory, OK '
	ENDIF
	! use of fftw needs creation of plans: 
	CALL fftinit( omeg(1:Nx,t_p*ydiff+1), omeghat(1:Ny,t_p*xdiff+1), Nx, Ny, PLAN1X, PLAN1Y, PLAN2X, PLAN2Y )

	IF (t_p.EQ.0) THEN
		PRINT *,'set up fft, OK '
	ENDIF

!	Initialisation of variables, initial data ....
	CALL init( Nx, Ny, n_p, t_p, xdiff, ydiff, info_p, &
			PLAN1X, PLAN1Y, PLAN2X, PLAN2Y, Lx, Ly, pi, &
			lowxv, upxv, lowyv, upyv, &
			Ky, Kx, Ky2, Kx2, omeg, omeghat, T1, T2, ffttemp, ffttemphat )
	IF (t_p.EQ.0) THEN
	   PRINT *,'set up initial data '
	ENDIF

!	save initial data			
	CALL compandsave( Nx, Ny, n_p, t_p, xdiff, ydiff, info_p, &
			 	plotnum, thefileu, thefilev, time, &
				lowxv, upxv, lowyv, upyv, omeg, omeghat, &
				PLOTARRAYU, PLOTARRAYV)
!	end of the initialisation
	
	IF (t_p.EQ.0) THEN
		PRINT *,'saved initial data '
	ENDIF
	!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
	!get initial nonlinear term using omeghat to find u and v!
	!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
	
	!u
	DO i=t_p*xdiff+1,(t_p+1)*xdiff
		DO j=1,Ny
			temp1(j,i)= cmplx(0.0d0,1.0d0)*ky(j)*omeghat(j,i)/(kx(i)*kx(i)+ky2(j)*ky2(j))
		END DO
	END DO				
 	
	CALL fft2dBack(Nx, Ny , n_p, t_p, xdiff, ydiff, info_p, &
						PLAN2X, PLAN2Y, lowxv, upxv, lowyv, upyv, &
						temp1, temp3, T1, T2, ffttemphat, ffttemp ) 

	!omega_x
 	DO i=t_p*xdiff+1,(t_p+1)*xdiff
		DO j=1,Ny
			temp1(j,i)=omeghat(j,i)*kx(i)*cmplx(0.0d0,1.0d0)
		END DO
	END DO
	
	CALL fft2dBack(Nx, Ny , n_p, t_p, xdiff, ydiff, info_p, &
					 PLAN2X, PLAN2Y, lowxv, upxv, lowyv, upyv, &
						temp1, temp2, T1, T2, ffttemphat, ffttemp ) 
	
 	DO j=t_p*ydiff+1,(t_p+1)*ydiff
		DO i=1,Nx
			nl(i,j)=temp3(i,j)*temp2(i,j)
		END DO
	END DO
	
 	!v
 	DO i=t_p*xdiff+1,(t_p+1)*xdiff
		DO j=1,Ny
			temp1(j,i)=kx(i)*omeghat(j,i)*cmplx(0.0d0,-1.0d0)&
				/(kx2(i)*kx2(i)+ky(j)*ky(j))
		END DO
	END DO		

	CALL fft2dBack(Nx, Ny , n_p, t_p, xdiff, ydiff, info_p, &
						PLAN2X, PLAN2Y, lowxv, upxv, lowyv, upyv, &
						temp1, temp3, T1, T2, ffttemphat, ffttemp ) 

	!omega_y
 	DO i=t_p*xdiff+1,(t_p+1)*xdiff
		DO j=1,Ny
			temp1(j,i)=omeghat(j,i)*ky(j)*cmplx(0.0d0,1.0d0)
		END DO
	END DO

	CALL fft2dBack(Nx, Ny , n_p, t_p, xdiff, ydiff, info_p, &
						PLAN2X, PLAN2Y, lowxv, upxv, lowyv, upyv, &
						temp1, temp2, T1, T2, ffttemphat, ffttemp) 
	
	!combine to get full nonlinear term in real space
 	DO j=t_p*ydiff+1,(t_p+1)*ydiff
		DO i=1,Nx
			nl(i,j)=(nl(i,j)+temp3(i,j)*temp2(i,j))
		END DO
	END DO
	CALL fft2dFor(Nx, Ny , n_p, t_p, xdiff, ydiff, info_p, &
						PLAN1X, PLAN1Y, lowxv, upxv, lowyv, upyv, &
						nl, nlhat, T1, T2, ffttemp, ffttemphat) 
	
	!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
	
	temp2=omeg !omegacheck
	IF (t_p.EQ.0) THEN
		PRINT *,'setup first nonlinear term, starting timestepping '
	ENDIF
	
!	system clock
	call system_clock(start,count_rate)
!	Solve the equation     
!	Start of the loop 
	DO k = 1,Nt
		chg=1.0d0
		DO i=t_p*xdiff+1,(t_p+1)*xdiff
			DO j=1,Ny
				rhsfixhat(j,i)=(dtInv-0.5d0*ReInv*(kx(i)*kx(i)+ky(j)*ky(j)))&
								*omeghat(j,i)-0.5d0*nlhat(j,i)
			END DO
		END DO 
		DO WHILE (chg>tol)

			DO i=t_p*xdiff+1,(t_p+1)*xdiff
				DO j=1,Ny
					omeghat(j,i)=(rhsfixhat(j,i) - 0.5d0*nlhat(j,i))&
							/(dtInv+0.5d0*ReInv*(kx(i)*kx(i)+ky(j)*ky(j)))   
				END DO
			END DO 
			
			CALL fft2dBack(Nx, Ny , n_p, t_p, xdiff, ydiff, info_p, &
						PLAN2X, PLAN2Y, lowxv, upxv, lowyv, upyv, &
						omeghat, omeg, T1, T2, ffttemphat, ffttemp) 

!			! .. Check convergence of fixed point iterations ..
			mychg=0.0d0
			chg=0.0d0
			DO j=t_p*ydiff+1,(t_p+1)*ydiff
				DO i=1,Nx
					mychg=mychg+abs((omeg(i,j)-temp2(i,j))*(omeg(i,j)-temp2(i,j)))				
				END DO
			END DO
		
			CALL MPI_ALLREDUCE(mychg,chg,1,MPI_DOUBLE_PRECISION,MPI_SUM,MPI_COMM_WORLD,info_p)

			!!!!!!!!!!!!!!!!
			!nonlinear term!
			!!!!!!!!!!!!!!!!
			!u			
			DO i=t_p*xdiff+1,(t_p+1)*xdiff
				DO j=1,Ny
					temp1(j,i)=ky(j)*omeghat(j,i)*cmplx(0.0d0,1.0d0)&
						/(kx(i)*kx(i)+ky2(j)*ky2(j))
				END DO
			END DO				
			
			CALL fft2dBack(Nx, Ny , n_p, t_p, xdiff, ydiff, info_p, &
						PLAN2X, PLAN2Y, lowxv, upxv, lowyv, upyv, &
						temp1, temp3, T1, T2, ffttemphat, ffttemp ) 
			!omega_x	
			DO i=t_p*xdiff+1,(t_p+1)*xdiff
				DO j=1,Ny
					temp1(j,i)=omeghat(j,i)*kx(i)*cmplx(0.0d0,1.0d0)
				END DO
			END DO
			
			CALL fft2dBack(Nx, Ny , n_p, t_p, xdiff, ydiff, info_p, &
						PLAN2X, PLAN2Y, lowxv, upxv, lowyv, upyv, &
						temp1, temp2, T1, T2, ffttemphat, ffttemp ) 
				
			DO j=t_p*ydiff+1,(t_p+1)*ydiff
				DO i=1,Nx
					nl(i,j)=temp3(i,j)*temp2(i,j)
				END DO
			END DO
			!v
			DO i=t_p*xdiff+1,(t_p+1)*xdiff
				DO j=1,Ny
					temp1(j,i)=kx(i)*omeghat(j,i)*cmplx(0.0d0,-1.0d0)&
						/(kx2(i)*kx2(i)+ky(j)*ky(j))
				END DO
			END DO		
			
			CALL fft2dBack(Nx, Ny , n_p, t_p, xdiff, ydiff, info_p, &
						PLAN2X, PLAN2Y, lowxv, upxv, lowyv, upyv, &
						temp1, temp3, T1, T2, ffttemphat, ffttemp) 
			!omega_y
			DO i=t_p*xdiff+1,(t_p+1)*xdiff
				DO j=1,Ny
					temp1(j,i)=omeghat(j,i)*ky(j)*cmplx(0.0d0,1.0d0)
				END DO
			END DO
			
			CALL fft2dBack(Nx, Ny , n_p, t_p, xdiff, ydiff, info_p, &
						PLAN2X, PLAN2Y, lowxv, upxv, lowyv, upyv, &
						temp1, temp2, T1, T2, ffttemphat, ffttemp ) 
				
			!combine to get full nonlinear term in real space
			DO j=t_p*ydiff+1,(t_p+1)*ydiff
				DO i=1,Nx
					nl(i,j)=(nl(i,j)+temp3(i,j)*temp2(i,j))
				END DO
			END DO
			CALL fft2dFor(Nx, Ny , n_p, t_p, xdiff, ydiff, info_p, &
						PLAN1X, PLAN1Y, lowxv, upxv, lowyv, upyv, &
						nl, nlhat, T1, T2, ffttemp, ffttemphat ) 
			!!!!!!!!!!!!!!!!				
			temp2=omeg !save omegacheck

		END DO
				
		time=time+DT
		! save data if taken the number of timesteps required between plots
!		IF (MOD(k,plotgap).EQ.0) THEN
!			CALL endwritebinfile(PLOTARRAYV,info_p,n_p,t_p,lowxv(t_p),upxv(t_p),Ny,Nx,thefilev)
!			CALL endwritebinfile(PLOTARRAYU,info_p,n_p,t_p,lowyv(t_p),upyv(t_p),Nx,Ny,thefileu)
!			CALL compandsave(Nx, Ny, n_p, t_p, xdiff, ydiff, info_p, &
!				plotnum, thefileu, thefilev, time,&
!				lowxv, upxv, lowyv, upyv, omeg, omeghat, &
!				PLOTARRAYU, PLOTARRAYV)
!			IF(t_p.EQ.0) THEN
!				PRINT *,' Saved data at time ',time
!			ENDIF									
!		ENDIF
	END DO 
!	fin loop
	CALL MPI_BARRIER(MPI_COMM_WORLD, info_p)
	CALL system_clock(finish,count_rate)
!	finish writing data to disk	
	CALL endwritebinfile(PLOTARRAYV,info_p,n_p,t_p,lowxv(t_p),upxv(t_p),Ny,Nx,thefilev)
	CALL endwritebinfile(PLOTARRAYU,info_p,n_p,t_p,lowyv(t_p),upyv(t_p),Nx,Ny,thefileu)

	!calculate error in final solution
	DO j=t_p*ydiff+1,(t_p+1)*ydiff
		y = REAL(j-1,kind(0d0))/REAL(Ny,kind=8) 
		DO i=1,Nx
			x = REAL(i-1,kind(0d0))/REAL(Nx,kind=8) 
			PLOTARRAYU(i,j)=(abs( 4.0d0*pi*sin(2.0d0*pi*x)*&
				sin(2.0d0*pi*y)*exp(-(8.0d0*ReInv*pi**2)*time)&
				-omeg(i,j)))
		END DO
	END DO

	mychg=maxval(PLOTARRAYU)
	CALL MPI_ALLREDUCE(mychg,calcerr,1,MPI_DOUBLE_PRECISION,MPI_MAX,MPI_COMM_WORLD,info_p)
	IF(t_p.eq.0) THEN
		PRINT*,'Processor ',t_p,' took ',REAL(finish-start)/REAL(count_rate),' for execution'
		PRINT*,'There were ',n_p,' MPI processes'
		PRINT*,'The grid size was ', Nx,' by ',Ny
		PRINT*,'The final error is ',calcerr
	END IF
	
	DEALLOCATE(lowxv, upxv, lowyv, upyv, &
			Kx,  Ky, Kx2,  Ky2, omeg, nl, &
			temp2, temp3, omeghat, nlhat,  &
			rhsfixhat, temp1, T1, T2, &
			ffttemp, ffttemphat, &
			PLOTARRAYU,PLOTARRAYV,&
			stat=AllocateStatus   )	 
	IF (AllocateStatus .NE. 0) STOP 

	IF (t_p.EQ.0) THEN
		PRINT *,'Deallocation of CPU memory, OK '
	ENDIF

	CALL fftdestroy(PLAN1X, PLAN1Y, PLAN2X, PLAN2Y )
	CALL MPI_FINALIZE(info_p)

	END PROGRAM NS2d
