	PROGRAM NS2d
	!-------------------------------------------------------------------------------------
	! .. PURPOSE ..
	!
	!	NS2d solves the two dimensional Navier Stokes equations
	! 	on a Square Domain [0,1]x[0,1] using pseudo-spectral methods and
	! 	Crank-Nicolson timestepping. The numerical solution is compared to
	! 	the exact Taylor-Green Vortex Solution. 
	!
	! .. PARAMETERS .. INITIALIZED IN INPUTFILE
	!	Re			= Reynolds number
	!	Nx			= power of two, number of modes in the x direction
	!	Ny			= power of two, number of modes in the y direction
	!	Nt			= the number of timesteps
	!	plotgap		= the number of timesteps to take before plotting
	!	Lx			= definition of the periodic domain of computation in x direction
	!	Ly			= definition of the periodic domain of computation in y direction
	!	EPSI		= dispersion parameter, 0 < EPSI << 1
	!	DT			= the time step
	!
	! .. Scalars ..
	!	W1, W2, W3, W4		= coefficients of the methods, see REFERENCES
	!	t_p			= the number of the MPI process  
	!	info_p			= MPI error code
	!	n_p			= the total number of MPI processes
	!	i, k, j			= integers used for loops
	!	deb, fin, co		= variables used for the calcul of the CPU time
	!	AllocateStatus		= variable used to verify the allocation of arrays
	!	pi			= the number pi, pi=4*Arctan(1).
	!	plotnum			= number of plots saved
	!	thefileu		= handle for file to be saved to disk
	!	thefilev		= handle for file to be saved to disk
	!
	! .. Variables, 1-dimension .. 	
	!	xdiff			= Nx/n_p  Variable used for the indiciation 
	!				(and so for the partage of datas between processes) 
	!	ydiff			= Ny/n_p  Variable used for the indiciation
	!				(and so for the partage of datas between processes) 
	!	PLAN1X			= data structure containing all the information that 
	!				fftw needs in order to compute the 1D forward Fourier transform
	!				in the x direction
	!	PLAN2X			= data structure containing all the information that 
	!				fftw needs in order to compute the 1D backward Fourier transform
	!				in the x direction
	!	PLAN1Y			= data structure containing all the information that 
	!				fftw needs in order to compute the 1D forward Fourier transform
	!				in the y direction
	!	PLAN2Y			= data structure containing all the information that 
	!				fftw needs in order to compute the 1D backward Fourier transform
	!				in the y direction
	!	MASSCHECK		= test of the numerical precision provided by mass conservation
	!	MA0			= initial mass numerically calculated via FFT
	!				(in both spatial coordinates)
	!
	! .. Arrays 1d .. (Vectors)
	!	COMP1			= (/ W1, W1 /), variable used to compute 
	!				the time splitting method (see REFERENCES) in a single loop 
	!	COMP2			= (/ W2, 0 /), variable used to compute 
	!				the time splitting method (see REFERENCES) in a single loop 
	!	lowxv, upxv, lowyv, upyv = Variables used for the indication, size: n_p
	!				low=lower, up=upper, xv= x value, yv=y value
	!	Kx			= Fourier frequencies in the x direction, size: lowxv(t_p):uppxv(t_p)  
	!	Kxb			= Fourier frequencies in the x direction, modified to avoid division by zero,
	!				size: lowxv(t_p):uppxv(t_p)  
	!	Ky			= Fourier frequencies in the y direction, size: Ny	
	!
	! .. Arrays 2d .. (Matrix)
	!	Note: Sizes are indicated without parallelisation
	!
	!	UINI			= u, solution of DS, size: [Nx Ny/n_p]
	!	VINI			= FFT(u), (FFT in both spatial coordinates), size: [Ny Nx/n_p]
	!	AUINI			= Absolute Value of u, squared; size: [Nx Ny/n_p]
	!	U2B			= -((abs(u).^2) + IFFT(K.*FFT(abs(u).^2))); SIZE: [Nx Ny/n_p]
	!				where, IFFT denotes the backward FFT in both spatial coordinates,
	!				and K denotes the operator with which we solve the second equation in
	!				DS, the equation for \phi. (K=(-2*Kx.^2)./(Kx.^2 + ALPHA*Ky.^2).
	!				Also used as the array to store data to be saved to disk
	! .. REFERENCES ..
	!
	! .. ACCURACY ..
	!		
	! .. ERROR INDICATORS AND WARNINGS ..
	! One might possibly use derived datatypes, MPI_ALLTOALL_W and/or change from 
	! storing data in 2d arrays to storing data in large 1d vectors with clever indexing to 
	! reduce memory requirements
	!
	! FURTHER COMMENTS
	! This program is based on a program for the Davey Stewartson equation originally by
	! B.K. Muite and K. Roidot
	!---------------------------------------------------------------------------------------------
	! EXTERNAL ROUTINES REQUIRED
	! MPI library
	! parainit 	-- intialize parameters, which are stored in the Input File 
	! fftinit	-- initializes fftw3 plans
	! writebinfile  -- saves a two dimensional field in native binary, double format using MPIIO
	! fft2dFor  	-- Computes the 2d-FFT forward in both spatial coordinates, by using 
	!		MPI_ALL_TO_ALL to perform the data transposition. Data is only transposed once, so
	!		output is in the opposite orientation to input.  Input data is destroyed.
	! fft2dBack	-- Computes the 2d-FFT backward in both spatial coordinates, by using 
	!		MPI_ALL_TO_ALL to perform the data transposition. Data is only transposed once, so
	!		output is in the opposite orientation to input.  Input data is destroyed.
	! compandsave	-- save the Fourier coefficients and the numerical solution
	!		in a binary file, by calling the subroutine writebinfile,
	!		and calculate the test for the mass conservation.
	! fftdestroy	-- destroys fftw arrays
	! init		-- initializes the grid points of computation, fourier frequencies, 
	!		the initial data, uini, and initial mass, and
	!		also saves initial data required to calculate points.
	!-----------------------------------------------------------------------------------------------------
	USE mpi
      
	IMPLICIT NONE    
!	Declaration of variables and parameter
	INTEGER(KIND=4)				::  Nx=1024, Ny=1024  
	INTEGER(KIND=4)				::  Nt=20, plotgap=1
	INTEGER(KIND=4)				::  i, k, j, start, finish, count_rate, AllocateStatus, nn, proc
	INTEGER(KIND=4)				::  info_p, n_p, t_p, xdiff, ydiff, plotnum, thefileu, thefilev
	REAL(KIND=8)				::  Re=1.0d0		
	REAL(KIND=8)				::	ReInv
	REAL(KIND=8), PARAMETER		::	tol=0.1d0**10
	REAL(KIND=8)				::  chg, mychg
	REAL(KIND=8)				::  time
	REAL(KIND=8)				::  x, y
	REAL(KIND=8), PARAMETER		::  pi= 3.14159265358979323846264338327950288419716939937510
	REAL(KIND=8)				::  dt=0.000125d0
	REAL(KIND=8)				::  dtInv
	REAL(KIND=8)				::  calcerr   
	INTEGER(KIND=8)									:: 	PLAN1X, PLAN1Y, PLAN2X, PLAN2Y
	INTEGER, DIMENSION(MPI_STATUS_SIZE) 			::  mpistatus
	INTEGER(KIND=4), DIMENSION(MPI_STATUS_SIZE,2*n_p)		:: mpistatusreq
	INTEGER(KIND=4), DIMENSION(:), ALLOCATABLE		::  lowxv, upxv, lowyv, upyv
	INTEGER(KIND=4), DIMENSION(:), ALLOCATABLE		::  reqAfor,reqAback,reqBfor,reqBback
	REAL(KIND=8), DIMENSION(:), ALLOCATABLE			::  Kx_d, Ky_d, Ky2_d, Kx2_d 
	COMPLEX(KIND=8), DIMENSION(:), ALLOCATABLE	 	::  T1A_d,T2A_d,T1B_d,T2B_d
	COMPLEX(KIND=8), DIMENSION(:), ALLOCATABLE		::  omeg_d,omeghat_d,nl_d, &
															rhsfixhat_d,omegx_d,omegy_d,u_d,v_d, &
															nlhat_d,temp1_d
	REAL(KIND=8), DIMENSION(:,:), ALLOCATABLE		::  PLOTARRAYU_d, PLOTARRAYV_d
!       Initialisation of data 
!       Initialisation MPI environment
	CALL MPI_INIT(info_p)
	CALL MPI_COMM_SIZE(MPI_COMM_WORLD, n_p, info_p)
	CALL MPI_COMM_RANK(MPI_COMM_WORLD, t_p, info_p)   
!       give a number to my proc from 0 to n_p-1

!	Initialisation of parameter
	CALL parainit(	Nx, Ny, Re, Nt, plotgap, &
			DT, time, t_p, n_p, info_p )

	dtInv=1.0d0/REAL(dt,kind(0d0))
	ReInv=1.0d0/REAL(Re,kind(0d0))
	plotnum=0
!	indiciation of array bounds
	ydiff = Ny/n_p
	xdiff = Nx/n_p

	ALLOCATE( lowxv(0:n_p-1) , lowyv(0:n_p-1), upxv(0:n_p-1), upyv(0:n_p-1), stat=AllocateStatus)
	IF (AllocateStatus .NE. 0) STOP

	IF (t_p.EQ.0) THEN
		PRINT *,'First allocation of CPU memory, OK '
	ENDIF
	
	! Calculate array indices
	DO i = 1,n_p
		lowxv(i-1) = (i-1)*xdiff + 1
		lowyv(i-1) = (i-1)*ydiff + 1
		upxv(i-1) = i*xdiff 
		upyv(i-1) = i*ydiff
	END DO

	IF (t_p.EQ.0) THEN
		PRINT *,'initialization of CPU array entires, OK '
	ENDIF
	
!	end indiciation 


	ALLOCATE(Kx_d(t_p*xdiff+1:(t_p+1)*xdiff),Ky_d(1:Ny),&
			 Kx2_d(t_p*xdiff+1:(t_p+1)*xdiff),Ky2_d(1:Ny),&
			 omeg_d(1:ydiff*Nx), nl_d(1:ydiff*Nx), &
			 omeghat_d(1:ydiff*Nx), nlhat_d(1:ydiff*Nx), & 
			 rhsfixhat_d(1:ydiff*Nx), &
			 omegx_d(1:ydiff*Nx), omegy_d(1:ydiff*Nx), &
			 u_d(1:ydiff*Nx), v_d(1:ydiff*Nx), &
			 temp1_d(1:ydiff*Nx), &
			 T1A_d(1:ydiff*Nx), T2A_d(1:ydiff*Nx), &
			 T1B_d(1:ydiff*Nx), T2B_d(1:ydiff*Nx), &
			 PLOTARRAYU_d(1:Nx,t_p*ydiff+1:(t_p+1)*ydiff),&
			 PLOTARRAYV_d(1:Ny,t_p*xdiff+1:(t_p+1)*xdiff),&
			 reqAfor(1:2*n_p),reqAback(1:2*n_p),&
			 reqBfor(1:2*n_p),reqBback(1:2*n_p),&	
			 stat=AllocateStatus   )	 
	IF (AllocateStatus .NE. 0) STOP 

	IF (t_p.EQ.0) THEN
		PRINT *,'allocation of memory, OK '
	ENDIF
	
	! Setup persistent communications requests
	
	DO nn=1,n_p
		proc=mod(t_p+nn,n_p)	! processor to or receive from
		CALL MPI_RECV_INIT(T2A_d(proc*xdiff*ydiff+1),xdiff*ydiff,&		
						  		MPI_DOUBLE_COMPLEX,proc,proc*n_p+t_p+n_p*n_p,&
						  		MPI_COMM_WORLD,reqAfor(nn),info_p)
		CALL MPI_SEND_INIT(T1A_d(proc*xdiff*ydiff+1),xdiff*ydiff,&
						 		MPI_DOUBLE_COMPLEX,proc,t_p*n_p+proc+n_p*n_p,&
								MPI_COMM_WORLD,reqAfor(nn+n_p),info_p)
	END DO
	DO nn=1,n_p
		proc=mod(t_p+nn,n_p)	! processor to or receive from
		CALL MPI_RECV_INIT(T1A_d(proc*xdiff*ydiff+1),xdiff*ydiff,&		
						  		MPI_DOUBLE_COMPLEX,proc,proc*n_p+t_p+n_p*n_p,&
						  		MPI_COMM_WORLD,reqAback(nn),info_p)
		CALL MPI_SEND_INIT(T2A_d(proc*xdiff*ydiff+1),xdiff*ydiff,&
						 		MPI_DOUBLE_COMPLEX,proc,t_p*n_p+proc+n_p*n_p,&
								MPI_COMM_WORLD,reqAback(nn+n_p),info_p)
	END DO
	DO nn=1,n_p
		proc=mod(t_p+nn,n_p)	! processor to or receive from
		CALL MPI_RECV_INIT(T2B_d(proc*xdiff*ydiff+1),xdiff*ydiff,&		
						  		MPI_DOUBLE_COMPLEX,proc,proc*n_p+t_p+n_p*n_p,&
						  		MPI_COMM_WORLD,reqBfor(nn),info_p)
		CALL MPI_SEND_INIT(T1B_d(proc*xdiff*ydiff+1),xdiff*ydiff,&
						 		MPI_DOUBLE_COMPLEX,proc,t_p*n_p+proc+n_p*n_p,&
								MPI_COMM_WORLD,reqBfor(nn+n_p),info_p)
	END DO
	DO nn=1,n_p
		proc=mod(t_p+nn,n_p)	! processor to or receive from
		CALL MPI_RECV_INIT(T1B_d(proc*xdiff*ydiff+1),xdiff*ydiff,&		
						  		MPI_DOUBLE_COMPLEX,proc,proc*n_p+t_p+n_p*n_p,&
						  		MPI_COMM_WORLD,reqBback(nn),info_p)
		CALL MPI_SEND_INIT(T2B_d(proc*xdiff*ydiff+1),xdiff*ydiff,&
						 		MPI_DOUBLE_COMPLEX,proc,t_p*n_p+proc+n_p*n_p,&
								MPI_COMM_WORLD,reqBback(nn+n_p),info_p)
	END DO

	! use of fftw needs creation of plans: 
	CALL fftinit( Nx, Ny, xdiff, ydiff, PLAN1X, PLAN1Y, PLAN2X, PLAN2Y )
	IF (t_p.EQ.0) THEN
		PRINT *,'set up fft, OK '
	ENDIF

!	Initialisation of variables, initial data ....
	CALL init(	Nx, Ny, n_p, t_p, xdiff, ydiff, info_p, &
			PLAN1X, PLAN1Y, pi, lowxv, upxv, lowyv, upyv ,&
			Ky_d, Kx_d,Ky2_d, Kx2_d,omeg_d )
	IF (t_p.EQ.0) THEN
	   PRINT *,'set up initial data '
	ENDIF

	CALL fft2dForStart(Nx,Ny,n_p,t_p,xdiff,ydiff,info_p,PLAN1X,omeg_d,T1A_d,T2A_d)	
	!CALL fft2dForTranspose(Nx,Ny,n_p,t_p,xdiff,ydiff,info_p,T1A_d,T2A_d)
	CALL MPI_STARTALL(2*n_p,reqAfor,info_p)
	CALL MPI_WAITALL(2*n_p,reqAfor,mpistatusreq,info_p)
	CALL fft2dForEnd(Nx,Ny,n_p,t_p,xdiff,ydiff,info_p,PLAN1Y,omeghat_d,T1A_d,T2A_d)
	CALL MPI_BARRIER(MPI_COMM_WORLD, info_p)
!	save initial data			
!	CALL compandsave(	Nx, Ny, n_p, t_p, xdiff, ydiff, info_p, &
!				 plotnum, thefileu, thefilev, time,&
!				lowxv, upxv, lowyv, upyv, omeg_d, omeghat_d, &
!				PLOTARRAYU_d, PLOTARRAYV_d)
!	end of the initialisation
	
	IF (t_p.EQ.0) THEN
		PRINT *,'saved initial data '
	ENDIF
	!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
	!get initial nonlinear term using omeghat to find u and v!
	!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
	
	!u
	DO i=t_p*xdiff+1,(t_p+1)*xdiff
		DO j=1,Ny
			temp1_d(Ny*(i-(t_p*xdiff+1))+j)= &
				cmplx(0.0d0,1.0d0)*ky_d(j)*omeghat_d(Ny*(i-(t_p*xdiff+1))+j)&
				/(kx_d(i)*kx_d(i)+ky2_d(j)*ky2_d(j))
		END DO
	END DO				
 	
	CALL fft2dBackStart(Nx,Ny,n_p,t_p,xdiff,ydiff,info_p,PLAN2Y,temp1_d,T1A_d,T2A_d)
	CALL MPI_STARTALL(2*n_p,reqAback,info_p)
	CALL MPI_WAITALL(2*n_p,reqAback,mpistatusreq,info_p)
	CALL fft2dBackEnd(Nx,Ny,n_p,t_p,xdiff,ydiff,info_p,PLAN2X,u_d,T1A_d,T2A_d)  

	!omega_x
 	DO i=t_p*xdiff+1,(t_p+1)*xdiff
		DO j=1,Ny
			temp1_d(Ny*(i-(t_p*xdiff+1))+j)=&
				omeghat_d(Ny*(i-(t_p*xdiff+1))+j)*kx_d(i)*cmplx(0.0d0,1.0d0)
		END DO
	END DO
	
	CALL fft2dBackStart(Nx,Ny,n_p,t_p,xdiff,ydiff,info_p,PLAN2Y,temp1_d,T1A_d,T2A_d)
	CALL MPI_STARTALL(2*n_p,reqAback,info_p)
	CALL MPI_WAITALL(2*n_p,reqAback,mpistatusreq,info_p)
	CALL fft2dBackEnd(Nx,Ny,n_p,t_p,xdiff,ydiff,info_p,PLAN2X,omegx_d,T1A_d,T2A_d)  
		
 	!v
	DO i=t_p*xdiff+1,(t_p+1)*xdiff
		DO j=1,Ny
			temp1_d(Ny*(i-(t_p*xdiff+1))+j)=&
				kx_d(i)*omeghat_d(Ny*(i-(t_p*xdiff+1))+j)*cmplx(0.0d0,-1.0d0)&
				/(kx2_d(i)*kx2_d(i)+ky_d(j)*ky_d(j))
		END DO
	END DO		

	CALL fft2dBackStart(Nx,Ny,n_p,t_p,xdiff,ydiff,info_p,PLAN2Y,temp1_d,T1A_d,T2A_d)
	CALL MPI_STARTALL(2*n_p,reqAback,info_p)
	CALL MPI_WAITALL(2*n_p,reqAback,mpistatusreq,info_p)
	CALL fft2dBackEnd(Nx,Ny,n_p,t_p,xdiff,ydiff,info_p,PLAN2X,v_d,T1A_d,T2A_d)  

	!omega_y
 	DO i=t_p*xdiff+1,(t_p+1)*xdiff
		DO j=1,Ny
			temp1_d(Ny*(i-(t_p*xdiff+1))+j)=&
				omeghat_d(Ny*(i-(t_p*xdiff+1))+j)*ky_d(j)*cmplx(0.0d0,1.0d0)
		END DO
	END DO

	CALL fft2dBackStart(Nx,Ny,n_p,t_p,xdiff,ydiff,info_p,PLAN2Y,temp1_d,T1A_d,T2A_d)
	CALL MPI_STARTALL(2*n_p,reqAback,info_p)
	CALL MPI_WAITALL(2*n_p,reqAback,mpistatusreq,info_p)
	CALL fft2dBackEnd(Nx,Ny,n_p,t_p,xdiff,ydiff,info_p,PLAN2X,omegy_d,T1A_d,T2A_d)  
	
	!combine to get full nonlinear term in real space
 	DO j=t_p*ydiff+1,(t_p+1)*ydiff
		DO i=1,Nx
			nl_d(Nx*(j-(t_p*ydiff+1))+i)=&
				(u_d(Nx*(j-(t_p*ydiff+1))+i)*omegx_d(Nx*(j-(t_p*ydiff+1))+i)&
				+v_d(Nx*(j-(t_p*ydiff+1))+i)*omegy_d(Nx*(j-(t_p*ydiff+1))+i))
		END DO
	END DO
	CALL fft2dForStart(Nx,Ny,n_p,t_p,xdiff,ydiff,info_p,PLAN1X,nl_d,T1A_d,T2A_d)	
	CALL MPI_STARTALL(2*n_p,reqAfor,info_p)
	CALL MPI_WAITALL(2*n_p,reqAfor,mpistatusreq,info_p)
	CALL fft2dForEnd(Nx,Ny,n_p,t_p,xdiff,ydiff,info_p,PLAN1Y,nlhat_d,T1A_d,T2A_d)
	
	
	IF (t_p.EQ.0) THEN
		PRINT *,'setup first nonlinear term, starting timestepping '
	ENDIF
!	system clock
	call system_clock(start,count_rate)
!	Solve the equation     
!	Start of the loop 
	DO k = 1,Nt
		chg=1.0d0
		DO i=t_p*xdiff+1,(t_p+1)*xdiff
			DO j=1,Ny
				rhsfixhat_d(Ny*(i-(t_p*xdiff+1))+j)=&
					(dtInv-0.5d0*ReInv*(kx_d(i)*kx_d(i)+ky_d(j)*ky_d(j)))&
					*omeghat_d(Ny*(i-(t_p*xdiff+1))+j)-0.5d0*nlhat_d(Ny*(i-(t_p*xdiff+1))+j)
			END DO
		END DO 
		DO WHILE (chg>tol)
			!omega_x	
			DO i=t_p*xdiff+1,(t_p+1)*xdiff
				DO j=1,Ny
					temp1_d(Ny*(i-(t_p*xdiff+1))+j)=&
						omeghat_d(Ny*(i-(t_p*xdiff+1))+j)*kx_d(i)*cmplx(0.0d0,1.0d0)
				END DO
			END DO
			CALL fft2dBackStart(Nx,Ny,n_p,t_p,xdiff,ydiff,info_p,PLAN2Y,temp1_d,T1A_d,T2A_d)
			CALL MPI_STARTALL(2*n_p,reqAback,info_p)
			!omega_y
			DO i=t_p*xdiff+1,(t_p+1)*xdiff
				DO j=1,Ny
					temp1_d(Ny*(i-(t_p*xdiff+1))+j)=&
						omeghat_d(Ny*(i-(t_p*xdiff+1))+j)*ky_d(j)*cmplx(0.0d0,1.0d0)
				END DO
			END DO
			CALL fft2dBackStart(Nx,Ny,n_p,t_p,xdiff,ydiff,info_p,PLAN2Y,temp1_d,T1B_d,T2B_d)
			CALL MPI_STARTALL(2*n_p,reqBback,info_p)
			CALL MPI_WAITALL(2*n_p,reqAback,mpistatusreq,info_p)
			CALL fft2dBackEnd(Nx,Ny,n_p,t_p,xdiff,ydiff,info_p,PLAN2X,omegx_d,T1A_d,T2A_d)  	
			DO i=t_p*xdiff+1,(t_p+1)*xdiff
				DO j=1,Ny
					omeghat_d(Ny*(i-(t_p*xdiff+1))+j)=&
						(rhsfixhat_d(Ny*(i-(t_p*xdiff+1))+j) - &
						 0.5d0*nlhat_d(Ny*(i-(t_p*xdiff+1))+j))&
						/(dtInv+0.5d0*ReInv*(kx_d(i)*kx_d(i)+ky_d(j)*ky_d(j)))   
				END DO
			END DO
			CALL MPI_WAITALL(2*n_p,reqBback,mpistatusreq,info_p)
			CALL fft2dBackEnd(Nx,Ny,n_p,t_p,xdiff,ydiff,info_p,PLAN2X,omegy_d,T1B_d,T2B_d)  
			!combine to get full nonlinear term in real space
			DO j=t_p*ydiff+1,(t_p+1)*ydiff
				DO i=1,Nx
					nl_d(Nx*(j-(t_p*ydiff+1))+i)=&
						(u_d(Nx*(j-(t_p*ydiff+1))+i)*omegx_d(Nx*(j-(t_p*ydiff+1))+i)+&
						 v_d(Nx*(j-(t_p*ydiff+1))+i)*omegy_d(Nx*(j-(t_p*ydiff+1))+i))
				END DO
			END DO
			CALL fft2dForStart(Nx,Ny,n_p,t_p,xdiff,ydiff,info_p,PLAN1X,nl_d,T1A_d,T2A_d)	
			CALL MPI_STARTALL(2*n_p,reqAfor,info_p)
			!u			
			DO i=t_p*xdiff+1,(t_p+1)*xdiff
				DO j=1,Ny
					temp1_d(Ny*(i-(t_p*xdiff+1))+j)=&
						ky_d(j)*omeghat_d(Ny*(i-(t_p*xdiff+1))+j)*cmplx(0.0d0,1.0d0)&
						/(kx_d(i)*kx_d(i)+ky2_d(j)*ky2_d(j))
				END DO
			END DO	
			CALL fft2dBackStart(Nx,Ny,n_p,t_p,xdiff,ydiff,info_p,PLAN2Y,temp1_d,T1B_d,T2B_d)
			CALL MPI_STARTALL(2*n_p,reqBback,info_p)
			CALL MPI_WAITALL(2*n_p,reqAfor,mpistatusreq,info_p)
			CALL fft2dForEnd(Nx,Ny,n_p,t_p,xdiff,ydiff,info_p,PLAN1Y,nlhat_d,T1A_d,T2A_d)			
			!v
			DO i=t_p*xdiff+1,(t_p+1)*xdiff
				DO j=1,Ny
					temp1_d(Ny*(i-(t_p*xdiff+1))+j)=&
						kx_d(i)*omeghat_d(Ny*(i-(t_p*xdiff+1))+j)*cmplx(0.0d0,-1.0d0)&
						/(kx2_d(i)*kx2_d(i)+ky_d(j)*ky_d(j))
				END DO
			END DO	
			CALL MPI_WAITALL(2*n_p,reqBback,mpistatusreq,info_p)
			CALL fft2dBackEnd(Nx,Ny,n_p,t_p,xdiff,ydiff,info_p,PLAN2X,u_d,T1B_d,T2B_d)  				
			CALL fft2dBackStart(Nx,Ny,n_p,t_p,xdiff,ydiff,info_p,PLAN2Y,temp1_d,T1A_d,T2A_d)	
			CALL MPI_STARTALL(2*n_p,reqAback,info_p)
			temp1_d=omeghat_d !save omegacheck
			DO i=t_p*xdiff+1,(t_p+1)*xdiff
				DO j=1,Ny
					omeghat_d(Ny*(i-(t_p*xdiff+1))+j)=&
						(rhsfixhat_d(Ny*(i-(t_p*xdiff+1))+j) - 0.5d0*nlhat_d(Ny*(i-(t_p*xdiff+1))+j))&
						/(dtInv+0.5d0*ReInv*(kx_d(i)*kx_d(i)+ky_d(j)*ky_d(j)))   
				END DO
			END DO 
			! .. Check convergence of fixed point iterations ..
			mychg=0.0d0
			chg=0.0d0
			DO i=t_p*xdiff+1,(t_p+1)*xdiff
				DO j=1,Ny
					mychg=mychg+(abs(	omeghat_d(Ny*(i-(t_p*xdiff+1))+j)-&
										temp1_d(Ny*(i-(t_p*xdiff+1))+j)))**2				
				END DO
			END DO
			CALL MPI_ALLREDUCE(mychg,chg,1,MPI_DOUBLE_PRECISION,MPI_SUM,MPI_COMM_WORLD,info_p)
			CALL MPI_WAITALL(2*n_p,reqAback,mpistatusreq,info_p)
			CALL fft2dBackEnd(Nx,Ny,n_p,t_p,xdiff,ydiff,info_p,PLAN2X,v_d,T1A_d,T2A_d) 
			!combine to get full nonlinear term in real space
			DO j=t_p*ydiff+1,(t_p+1)*ydiff
				DO i=1,Nx
					nl_d(Nx*(j-(t_p*ydiff+1))+i)=&
						(u_d(Nx*(j-(t_p*ydiff+1))+i)*omegx_d(Nx*(j-(t_p*ydiff+1))+i)+&
						 v_d(Nx*(j-(t_p*ydiff+1))+i)*omegy_d(Nx*(j-(t_p*ydiff+1))+i))
				END DO
			END DO	
			CALL fft2dForStart(Nx,Ny,n_p,t_p,xdiff,ydiff,info_p,PLAN1X,nl_d,T1B_d,T2B_d)	
			CALL MPI_STARTALL(2*n_p,reqBfor,info_p)
			CALL MPI_WAITALL(2*n_p,reqBfor,mpistatusreq,info_p)
			CALL fft2dForEnd(Nx,Ny,n_p,t_p,xdiff,ydiff,info_p,PLAN1Y,nlhat_d,T1B_d,T2B_d)
		END DO
		time=time+DT
		! save data if taken the number of timesteps required between plots
!		IF (MOD(k,plotgap).EQ.0) THEN
!			CALL fft2dBackStart(Nx,Ny,n_p,t_p,xdiff,ydiff,info_p,PLAN2Y,omeghat_d,T1A_d,T2A_d)
!			CALL fft2dBackTransposeSt(Nx,Ny,n_p,t_p,xdiff,ydiff,info_p,T1A_d,T2A_d) 
!			CALL MPI_STARTALL(2*n_p,reqAback,info_p)
!			CALL MPI_WAITALL(2*n_p,reqAback,mpistatusreq,info_p)
!			CALL fft2dBackEnd(Nx,Ny,n_p,t_p,xdiff,ydiff,info_p,PLAN2X,omeg_d,T1A_d,T2A_d)  
!
!			CALL endwritebinfile(PLOTARRAYV_d,info_p,n_p,t_p,lowxv(t_p),upxv(t_p),Ny,Nx,thefilev)
!			CALL endwritebinfile(PLOTARRAYU_d,info_p,n_p,t_p,lowyv(t_p),upyv(t_p),Nx,Ny,thefileu)
!			CALL compandsave(Nx, Ny, n_p, t_p, xdiff, ydiff, info_p, &
!				plotnum, thefileu, thefilev, time,&
!				lowxv, upxv, lowyv, upyv, omeg_d, omeghat_d, &
!				PLOTARRAYU_d, PLOTARRAYV_d)
!			IF(t_p.EQ.0) THEN
!				PRINT *,' Saved data at time ',time
!			ENDIF									
!		ENDIF
	END DO 
!	fin loop
	CALL MPI_BARRIER(MPI_COMM_WORLD, info_p)
	CALL system_clock(finish,count_rate)

	CALL fft2dBackStart(Nx,Ny,n_p,t_p,xdiff,ydiff,info_p,PLAN2Y,omeghat_d,T1A_d,T2A_d)
	CALL MPI_STARTALL(2*n_p,reqAback,info_p)
	CALL MPI_WAITALL(2*n_p,reqAback,mpistatusreq,info_p)
	CALL fft2dBackEnd(Nx,Ny,n_p,t_p,xdiff,ydiff,info_p,PLAN2X,omeg_d,T1A_d,T2A_d)  

!	finish writing data to disk	
!	CALL endwritebinfile(PLOTARRAYV_d,info_p,n_p,t_p,lowxv(t_p),upxv(t_p),Ny,Nx,thefilev)
!	CALL endwritebinfile(PLOTARRAYU_d,info_p,n_p,t_p,lowyv(t_p),upyv(t_p),Nx,Ny,thefileu)

	!calculate error in final solution
	DO j=t_p*ydiff+1,(t_p+1)*ydiff
		y = REAL(j-1,kind(0d0))/REAL(Ny,kind=8) 
		DO i=1,Nx
			x = REAL(i-1,kind(0d0))/REAL(Nx,kind=8) 
			PLOTARRAYU_d(i,j)=(abs( 4.0d0*pi*sin(2.0d0*pi*x)*&
				sin(2.0d0*pi*y)*exp(-(8.0d0*ReInv*pi**2)*time)&
				-omeg_d(Nx*(j-(t_p*ydiff+1))+i)))
		END DO
	END DO

	mychg=maxval(PLOTARRAYU_d)
	CALL MPI_ALLREDUCE(mychg,calcerr,1,MPI_DOUBLE_PRECISION,MPI_MAX,MPI_COMM_WORLD,info_p)
	IF(t_p.eq.0) THEN
		PRINT*,'Processor ',t_p,' took ',REAL(finish-start)/REAL(count_rate),' for execution'
		PRINT*,'There were ',n_p,' MPI processes'
		PRINT*,'The grid size was ', Nx,' by ',Ny
		PRINT*,'The final error is ',calcerr
	END IF
	
	DO nn=1,2*n_p					
		CALL MPI_REQUEST_FREE(reqAfor(nn),info_p)
		CALL MPI_REQUEST_FREE(reqAback(nn),info_p)
		CALL MPI_REQUEST_FREE(reqBfor(nn),info_p)
		CALL MPI_REQUEST_FREE(reqBback(nn),info_p)
	END DO
	
	DEALLOCATE(lowxv, upxv, lowyv, upyv, Kx_d,  Ky_d, &
			omeg_d, nl_d, omeghat_d, nlhat_d,  &
			rhsfixhat_d, omegx_d, omegy_d, &
			u_d, v_d, temp1_d, &
			T1A_d, T2A_d, T1B_d, T2B_d, &
			PLOTARRAYU_d,PLOTARRAYV_d,&
			reqAfor,reqAback,reqBfor,reqBback,&
			stat=AllocateStatus   )	 
	IF (AllocateStatus .NE. 0) STOP 

	IF (t_p.EQ.0) THEN
		PRINT *,'Deallocation of  memory, OK '
	ENDIF

	CALL fftdestroy(PLAN1X, PLAN1Y, PLAN2X, PLAN2Y)
	CALL MPI_FINALIZE(info_p)

	END PROGRAM NS2d
