!--------------------------------------------------------------------
!
!
! PURPOSE
!
! This subroutine program calculates the energy for the nonlinear
! sine-Gordon equation in 2 dimensions
! u_{tt}-u_{xx}+u_{yy}=-sin(u)
!
! The energy density is given by
! 0.5u_t^2+0.5u_x^2+0.5u_y^2+(1-cos(u))
!
! AUTHORS
!
! B. Cloutier, B.K. Muite, P. Rigge
! 4 June 2012
!
! INPUT
!
! .. Scalars ..
!  Nx                           = number of modes in x - power of 2 for FFT
!  Ny                           = number of modes in y - power of 2 for FFT
!  planfxy                      = Forward 2d fft plan
!  planbxy                      = Backward 2d fft plan
!  dt                           = timestep
! .. Arrays ..
!  u                            = approximate solution
!  v                            = Fourier transform of approximate solution
!  uold                         = approximate solution
!  vold                         = Fourier transform of approximate solution
!  temp1                        = array to hold temporary values
!  temp2                        = array to hold temporary values
! .. Vectors ..
!  kx                           = fourier frequencies in x direction
!  ky                           = fourier frequencies in y direction
!
! OUTPUT
!
! .. Scalars ..
!  enkin                        = Kinetic energy
!  enstr                        = Strain energy
!  enpot                        = Potential energy
!  en                           = Total energy
!
! LOCAL VARIABLES
!
! .. Scalars ..
!  j                            = loop counter in y direction
!
! FURTHER COMMENTS
! Check that the initial iterate is consistent with the
! boundary conditions for the domain specified
!--------------------------------------------------------------------
! External routines required
!
! External libraries required
!       FFTW3    -- Fast Fourier Transform in the West Library
!                       (http://www.fftw.org/)
SUBROUTINE enercalc(Nx,Ny,planfxy,planbxy,dt,enkin,enstr,&
     enpot,en,kx,ky,temp1,temp2,u,uold)
  IMPLICIT NONE
  INTEGER(KIND=4), INTENT(IN)                             	:: Nx,Ny
  REAL(KIND=8), INTENT(IN)                                	:: dt
  INTEGER(KIND=8), INTENT(IN)                             	:: planfxy
  INTEGER(KIND=8), INTENT(IN)                             	:: planbxy
  REAL(KIND=8), DIMENSION(1:Nx),INTENT(IN)             		:: kx
  REAL(KIND=8), DIMENSION(1:Ny),INTENT(IN)             		:: ky
  REAL   (KIND=8), DIMENSION(1:Nx,1:Ny),INTENT(IN)        	:: u,uold
  COMPLEX(KIND=8), DIMENSION(1:Nx,1:Ny),INTENT(INOUT)     	:: temp1,temp2
  REAL(KIND=8), INTENT(OUT)                               	:: enkin,enstr
  REAL(KIND=8), INTENT(OUT)                               	:: enpot,en
  INTEGER(KIND=4)                                         	:: i,j
  DO j=1,Ny
  	DO i=1,Nx
     	temp1(i,j)=0.5d0*(uold(i,j)+u(i,j))
     END DO
  END DO
  CALL dfftw_execute_dft_(planfxy,temp1,temp2)
  !.. Strain energy ..
  DO j=1,Ny
  	DO i=1,Nx
     temp1(i,j)=cmplx(0.0d0,1.0d0)*kx(i)*temp2(i,j)
	END DO
  END DO
  CALL dfftw_execute_dft_(planbxy,temp1,temp2)
  DO j=1,Ny
  	DO i=1,Nx
     temp2(i,j)=abs(temp2(i,j)/REAL(Nx*Ny,kind(0d0)))**2
     END DO
  END DO
  CALL dfftw_execute_dft_(planfxy,temp2,temp1)
  enstr=0.5d0*REAL(abs(temp1(1,1)),kind(0d0))/REAL(Nx*Ny,kind(0d0))
  DO j=1,Ny
	DO i=1,Nx
     	temp1(i,j)=0.5d0*(uold(i,j)+u(i,j))
	END DO
  END DO
  CALL dfftw_execute_dft_(planfxy,temp1,temp2)
  DO j=1,Ny
  	DO i=1,Nx
     temp1(i,j)=cmplx(0.0d0,1.0d0)*ky(j)*temp2(i,j)
    END DO
  END DO
  CALL dfftw_execute_dft_(planbxy,temp1,temp2)
  DO j=1,Ny
  	DO i=1,Nx
     temp2(i,j)=abs(temp2(i,j)/REAL(Nx*Ny,kind(0d0)))**2
    END DO
  END DO
  CALL dfftw_execute_dft_(planfxy,temp2,temp1)
  enstr=enstr+0.5d0*REAL(abs(temp1(1,1)),kind(0d0))/REAL(Nx*Ny,kind(0d0))
  ! .. Kinetic Energy ..
  DO j=1,Ny
  	DO i=1,Nx
     temp2(i,j)=( abs(u(i,j)-uold(i,j))/dt )**2
    END DO
  END DO
  CALL dfftw_execute_dft_(planfxy,temp2,temp1)
  enkin=0.5d0*REAL(abs(temp1(1,1)),kind(0d0))/REAL(Nx*Ny,kind(0d0))
  ! .. Potential Energy ..
  DO j=1,Ny
  	DO i=1,Nx
     temp2(i,j)=1-cos(0.5*u(i,j)+0.5*uold(i,j))
     END DO
  END DO
  CALL dfftw_execute_dft_(planfxy,temp2,temp1)
  enpot=REAL(abs(temp1(1,1)),kind(0d0))/REAL(Nx*Ny,kind(0d0))
  en=enpot+enkin+enstr
END SUBROUTINE enercalc
